# Licensed under a 3-clause BSD style license - see LICENSE.rst
# flake8: noqa: F403, F405

import pytest
import numpy as np
from astropop.polarimetry.dualbeam import match_pairs, estimate_dxdy, \
                                          _compute_theta, quarterwave_model, \
                                          halfwave_model, \
                                          _DualBeamPolarimetry, \
                                          SLSDualBeamPolarimetry, \
                                          StokesParameters
from astropop.math import QFloat
from astropy import units
from astropop.testing import *
from scipy.optimize import curve_fit
from functools import partial


def get_flux_oe(flux, psi, k, q, u, v=None, zero=0):
    """Get ordinary and extraordinary fluxes."""
    if v is None:
        zi = halfwave_model(psi, q, u, zero=zero)
    else:
        zi = quarterwave_model(psi, q, u, v, zero=zero)
    fo = flux*(1+zi)/2
    fe = flux*(1-zi)/2
    return fo, fe/k


class DummyPolarimeter(_DualBeamPolarimetry):
    def _half_fit(self, psi, zi):
        """Fit the Stokes params for halfwave retarder."""

    def _quarter_fit(self, psi, zi):
        """Fit the Stokes params for quarterwave retarder."""


def test_compute_theta():
    assert_equal(_compute_theta(0, 1), 45*units.degree)
    assert_equal(_compute_theta(1, 0), 0*units.degree)
    assert_equal(_compute_theta(0, -1), 135*units.degree)
    assert_equal(_compute_theta(-1, 0), 90*units.degree)
    assert_equal(_compute_theta(0.5, 0.5), 22.5*units.degree)
    assert_equal(_compute_theta(0.5, -0.5), 157.5*units.degree)
    assert_equal(_compute_theta(-0.5, 0.5), 67.5*units.degree)
    assert_equal(_compute_theta(-0.5, -0.5), 112.5*units.degree)


class Test_PairMatching:
    def test_estimate_dxdy(self):
        dx, dy = 30.5, -25.2
        x1 = np.random.uniform(low=0, high=1024, size=500)
        y1 = np.random.uniform(low=0, high=1024, size=500)
        x2 = x1 + dx
        y2 = y1 + dy

        x = np.concatenate((x1, x2))
        y = np.concatenate((y1, y2))

        dx_e, dy_e = estimate_dxdy(x, y, steps=[50, 5, 1], bins=100)

        # 1 decimal is enough
        assert_almost_equal(dx, dx_e, decimal=1)
        assert_almost_equal(dy, dy_e, decimal=1)

    def test_match_pairs(self):
        dx, dy = 30.5, -25.2

        x1 = np.array([46.7, 68.3, 131.9, 83.5, 34.7, 170.2, 96.0, 115.9,
                       155.6, 25.8])
        y1 = np.array([186.5, 29.9, 94.9, 105.2, 43.4, 108.8, 177.6, 18.5, 5.4,
                       119.6])
        x2 = x1 + dx
        y2 = y1 + dy

        x = np.concatenate((x1, x2))
        y = np.concatenate((y1, y2))

        index = match_pairs(x, y, dx, dy, 0.5)

        assert_equal(len(index), 10)
        assert_equal(index['o'], np.arange(10, 20))
        assert_equal(index['e'], np.arange(0, 10))

    def test_match_pairs_negative(self):
        dx, dy = 30.5, -25.2

        x1 = np.array([46.7, 68.3, 131.9, 83.5, 34.7, 170.2, 96.0, 115.9,
                       155.6, 25.8])
        y1 = np.array([186.5, 29.9, 94.9, 105.2, 43.4, 108.8, 177.6, 18.5, 5.4,
                       119.6])
        x2 = x1 - dx
        y2 = y1 - dy

        x = np.concatenate((x1, x2))
        y = np.concatenate((y1, y2))

        index = match_pairs(x, y, dx, dy, 0.5)

        assert_equal(len(index), 10)
        assert_equal(index['o'], np.arange(0, 10))
        assert_equal(index['e'], np.arange(10, 20))


class Test_ModelQuarter:
    def test_model_evaluate_plain(self):
        q = 0.0130
        u = -0.0021
        v = 0.03044
        zero = 60
        # values simulated that match IRAF pccdpack
        expect = np.array([-0.02220249, 0.00477571, 0.02406067, 0.02974862,
                           0.03052114, 0.02053262, -0.00637933, -0.02905695,
                           -0.02220249, 0.00477571, 0.02406067, 0.02974862,
                           0.03052114, 0.02053262, -0.00637933, -0.02905695])

        psi = np.arange(0, 360, 22.5)
        zi = quarterwave_model(psi, q=q, u=u, v=v, zero=zero)
        assert_almost_equal(zi, expect)

    def test_model_evaluate_quantity(self):
        q = 0.0130
        u = -0.0021
        v = 0.03044
        # values simulated that match IRAF pccdpack
        expect = np.array([-0.02220249, 0.00477571, 0.02406067, 0.02974862,
                           0.03052114, 0.02053262, -0.00637933, -0.02905695,
                           -0.02220249, 0.00477571, 0.02406067, 0.02974862,
                           0.03052114, 0.02053262, -0.00637933, -0.02905695])

        psi = np.arange(0, 360, 22.5)*units.degree
        zero = 60*units.degree
        zi = quarterwave_model(psi, q=q, u=u, v=v, zero=zero)
        assert_almost_equal(zi, expect)

        psi = np.arange(0, 2*np.pi, np.pi/8)*units.radian
        zero = 60*units.degree
        zi = quarterwave_model(psi, q=q, u=u, v=v, zero=zero)
        assert_almost_equal(zi, expect)

    def test_fit(self):
        q = 0.0130
        u = -0.0021
        v = 0.03044
        zero = 60
        psi = np.arange(0, 360.5, 22.5)
        zi = quarterwave_model(psi, q=q, u=u, v=v, zero=zero)

        model = partial(quarterwave_model, zero=zero)

        fit, cov = curve_fit(model, psi, zi,
                             bounds=([-1, -1, -1], [1, 1, 1]),
                             method='trf')
        assert_almost_equal(fit, [q, u, v], decimal=3)

    def test_fit_free_zero(self):
        q = 0.0130
        u = -0.0021
        v = 0.03044
        zero = 60
        psi = np.arange(0, 360, 22.5)
        zi = quarterwave_model(psi, q=q, u=u, v=v, zero=zero)

        fit, cov = curve_fit(quarterwave_model, psi, zi,
                             bounds=([-1, -1, -1, 0], [1, 1, 1, 180]),
                             method='trf')
        assert_almost_equal(fit, [q, u, v, zero], decimal=3)


class Test_ModelHalf:
    def test_model_evaluate_plain(self):
        q = 0.0130
        u = -0.021
        zero = 60
        expect = np.array([0.01168653, 0.02175833, -0.01168653, -0.02175833,
                           0.01168653, 0.02175833, -0.01168653, -0.02175833,
                           0.01168653, 0.02175833, -0.01168653, -0.02175833,
                           0.01168653, 0.02175833, -0.01168653, -0.02175833])

        psi = np.arange(0, 360, 22.5)
        zi = halfwave_model(psi, q=q, u=u, zero=zero)
        assert_almost_equal(zi, expect)

    def test_model_evaluate_quantity(self):
        q = 0.0130
        u = -0.021
        expect = np.array([0.01168653, 0.02175833, -0.01168653, -0.02175833,
                           0.01168653, 0.02175833, -0.01168653, -0.02175833,
                           0.01168653, 0.02175833, -0.01168653, -0.02175833,
                           0.01168653, 0.02175833, -0.01168653, -0.02175833])

        psi = np.arange(0, 360, 22.5)*units.degree
        zero = 60*units.degree
        zi = halfwave_model(psi, q=q, u=u, zero=zero)
        assert_almost_equal(zi, expect)

        psi = np.arange(0, 2*np.pi, np.pi/8)*units.radian
        zero = 60*units.degree
        zi = halfwave_model(psi, q=q, u=u, zero=zero)
        assert_almost_equal(zi, expect)

    def test_fit(self):
        q = 0.0130
        u = -0.021
        zero = 60
        psi = np.arange(0, 360.5, 22.5)
        zi = halfwave_model(psi, q=q, u=u, zero=zero)

        model = partial(halfwave_model, zero=zero)

        fit, cov = curve_fit(model, psi, zi,
                             bounds=([-1, -1], [1, 1]),
                             method='trf')
        assert_almost_equal(fit, [q, u], decimal=3)


class Test_DummyPolarimetry:
    @pytest.mark.parametrize('kwargs', [{}, {'zero': 60},
                                        {'zero': 60, 'k': 1.2},
                                        {'zero': 60, 'k': 1.2, 'min_snr': 80},
                                        {'k': 1.0}, {'min_snr': 80},
                                        {'compute_k': True}])
    def test_initialize_ok(self, kwargs):
        pol = DummyPolarimeter('halfwave', **kwargs)
        assert_equal(pol.retarder, 'halfwave')
        if 'k' in kwargs:
            assert_equal(pol.k, kwargs['k'])
        else:
            assert_is_none(pol.k)
        if 'zero' in kwargs:
            assert_equal(pol.zero, kwargs['zero'])
        else:
            assert_is_none(pol.zero)
        if 'min_snr' in kwargs:
            assert_equal(pol.min_snr, kwargs['min_snr'])
        else:
            assert_is_none(pol.min_snr)
        if 'compute_k' in kwargs:
            assert_equal(pol.compute_k, kwargs['compute_k'])
        else:
            assert_false(pol.compute_k)

    def test_initialize_error_redundancy(self):
        with pytest.raises(ValueError, match='k and compute_k cannot be used '
                           'together.'):
            DummyPolarimeter('halfwave', k=1.2, compute_k=True)

    def test_initialize_error_retarder(self):
        with pytest.raises(ValueError, match="Retarder dummy unknown."):
            DummyPolarimeter('dummy')

    def test_initialize_zero_quantity(self):
        pol = DummyPolarimeter('halfwave', zero=60*units.degree)
        assert_equal(pol.zero, 60)

    def test_initialize_n_pos(self):
        pol = DummyPolarimeter('halfwave')
        assert_equal(pol._n_pos, 4)
        pol = DummyPolarimeter('quarterwave')
        assert_equal(pol._n_pos, 8)

    def test_estimate_normalize_half_ok(self):
        pol = DummyPolarimeter('halfwave')
        psi = np.arange(0, 360, 22.5)
        flux_o, flux_e = get_flux_oe(1e5, psi, k=1.2, q=0.0130,
                                     u=-0.021, zero=60)
        k = pol._estimate_normalize_half(psi, flux_o, flux_e)
        assert_almost_equal(k, 1.2)

    def test_estimate_normalize_half_error(self):
        pol = DummyPolarimeter('halfwave')
        psi = np.arange(0, 360, 22.5)

        flux_o, flux_e = get_flux_oe(1e5, psi, k=1.2, q=0.0130,
                                     u=-0.021, zero=60)
        flux_o[np.where(psi // 22.5 == 0)] = np.nan
        with pytest.raises(ValueError, match='Could not estimate the '
                           'normalization factor.'):
            pol._estimate_normalize_half(psi, flux_o, flux_e)

        flux_o, flux_e = get_flux_oe(1e5, psi, k=1.2, q=0.0130,
                                     u=-0.021, zero=60)
        flux_e[np.where(psi // 22.5 == 0)] = np.nan
        with pytest.raises(ValueError, match='Could not estimate the '
                           'normalization factor.'):
            pol._estimate_normalize_half(psi, flux_o, flux_e)

    def test_estimate_normalize_quarter_ok(self):
        q = 0.0130
        pol = DummyPolarimeter('halfwave')
        psi = np.arange(0, 360, 22.5)
        flux_o, flux_e = get_flux_oe(1e5, psi, k=1.2, q=0.0130,
                                     u=-0.021, zero=60)

        pol = DummyPolarimeter('quarterwave')
        k = pol._estimate_normalize_quarter(psi, flux_o, flux_e, q)
        assert_almost_equal(k,
                            np.sum(flux_o)/np.sum(flux_e)*(1-0.5*q)/(1+0.5*q))

    def test_check_positions(self):
        pol = DummyPolarimeter('halfwave')
        pol._check_positions(np.arange(0, 360, 22.5))

        psi = np.arange(0, 360, 22.5)
        psi[2] = 45.2
        with pytest.raises(ValueError, match="Retarder positions must be "
                           "multiple of 22.5 deg"):
            pol._check_positions(psi)

    def test_calc_zi(self):
        pol = DummyPolarimeter('halfwave')
        psi = np.arange(0, 360, 22.5)
        flux_o, flux_e = get_flux_oe(1e5, psi, k=1.2, q=0.0130,
                                     u=-0.021, zero=60)
        expect = halfwave_model(psi, q=0.0130, u=-0.021, zero=60)
        zi = pol._calc_zi(flux_o, flux_e, k=1.2)
        assert_almost_equal(zi, expect)


class Test_StokesParameters:
    def test_initialize_halfwave(self):
        q = QFloat(0.0130, 0.001)
        u = QFloat(-0.021, 0.001)

        # V is not needed
        pol = StokesParameters(retarder='halfwave', q=q, u=u)
        assert_equal(pol.q, q)
        assert_equal(pol.u, u)
        assert_is_none(pol.v)
        assert_equal(pol.retarder, 'halfwave')
        assert_equal(pol.k, 1.0)
        assert_equal(pol.zero, 0.0)

    def test_initialize_halfwave_kzero(self):
        q = QFloat(0.0130, 0.001)
        u = QFloat(-0.021, 0.001)
        k = QFloat(1.2, 0.001)
        zero = QFloat(60, 0.001)

        # V is not needed
        pol = StokesParameters(retarder='halfwave', q=q, u=u, k=k, zero=zero)
        assert_equal(pol.q, q)
        assert_equal(pol.u, u)
        assert_is_none(pol.v)
        assert_equal(pol.retarder, 'halfwave')
        assert_equal(pol.k, k)
        assert_equal(pol.zero, zero)

    def test_initialize_quarterwave(self):
        q = QFloat(0.0130, 0.001)
        u = QFloat(-0.021, 0.001)
        v = QFloat(0.021, 0.001)

        # V is needed
        pol = StokesParameters(retarder='quarterwave', q=q, u=u, v=v)
        assert_equal(pol.q, q)
        assert_equal(pol.u, u)
        assert_equal(pol.v, v)
        assert_equal(pol.retarder, 'quarterwave')
        assert_equal(pol.k, 1.0)
        assert_equal(pol.zero, 0.0)

    def test_initialize_quarterwave_kzero(self):
        q = QFloat(0.0130, 0.001)
        u = QFloat(-0.021, 0.001)
        v = QFloat(0.021, 0.001)
        k = QFloat(1.2, 0.001)
        zero = QFloat(60, 0.001)

        # V is needed
        pol = StokesParameters(retarder='quarterwave', q=q, u=u, v=v, k=k,
                               zero=zero)
        assert_equal(pol.q, q)
        assert_equal(pol.u, u)
        assert_equal(pol.v, v)
        assert_equal(pol.retarder, 'quarterwave')
        assert_equal(pol.k, k)
        assert_equal(pol.zero, zero)

    def test_stokes_properties(self):
        q = QFloat(0.0130, 0.001)
        u = QFloat(-0.021, 0.001)
        v = QFloat(0.021, 0.001)
        pol = StokesParameters(retarder='quarterwave', q=q, u=u, v=v)

        assert_almost_equal(pol.p.nominal, 0.0246981, decimal=3)
        assert_almost_equal(pol.theta.nominal, 150.8797, decimal=3)


class Test_SLSPolarimetry:
    def test_fit_half(self):
        q = 0.0130
        u = -0.021
        psi = np.arange(0, 360, 22.5)
        flux_o, flux_e = get_flux_oe(1e5, psi, k=1.0, q=q, u=u, zero=0)
        pol = SLSDualBeamPolarimetry(retarder='halfwave', k=1.0)
        p = pol.compute(psi, flux_o, flux_e,
                        f_ord_error=[50]*16, f_ext_error=[50]*16)
        assert_almost_equal(p.q.nominal, q)
        assert_almost_equal(p.u.nominal, u)
        assert_almost_equal(p.k, 1.0)
        assert_is_none(p.zero)

    def test_fit_half_no_errors(self):
        q = 0.0130
        u = -0.021
        psi = np.arange(0, 360, 22.5)
        flux_o, flux_e = get_flux_oe(1e5, psi, k=1.0, q=q, u=u, zero=0)
        pol = SLSDualBeamPolarimetry(retarder='halfwave', k=1.0)
        p = pol.compute(psi, flux_o, flux_e)
        assert_almost_equal(p.q.nominal, q)
        assert_almost_equal(p.u.nominal, u)
        assert_almost_equal(p.k, 1.0)
        assert_is_none(p.zero)

    def test_fit_half_estimate_k(self):
        q = 0.0130
        u = -0.021
        psi = np.arange(0, 360, 22.5)
        flux_o, flux_e = get_flux_oe(1e5, psi, k=1.2, q=q, u=u, zero=0)
        pol = SLSDualBeamPolarimetry(retarder='halfwave', compute_k=True)
        p = pol.compute(psi, flux_o, flux_e,
                        f_ord_error=[50]*16, f_ext_error=[50]*16)
        assert_almost_equal(p.q.nominal, q)
        assert_almost_equal(p.u.nominal, u)
        assert_almost_equal(p.k.nominal, 1.2)
        assert_is_none(p.zero)

    def test_fit_quarter(self):
        q = 0.0130
        u = -0.027
        v = 0.021
        zero = 60

        psi = np.arange(0, 360, 22.5)
        flux_o, flux_e = get_flux_oe(1e5, psi, k=1.0, q=q, u=u, v=v, zero=zero)
        pol = SLSDualBeamPolarimetry(retarder='quarterwave', zero=60,
                                     compute_k=True)
        p = pol.compute(psi, flux_o, flux_e,
                        f_ord_error=[50]*16, f_ext_error=[50]*16)

        assert_almost_equal(p.q.nominal, q)
        assert_almost_equal(p.u.nominal, u)
        assert_almost_equal(p.v.nominal, v)
        assert_almost_equal(p.k.nominal, 1.0, decimal=2)
        assert_almost_equal(p.zero, zero)

    def test_fit_quarter_no_errors(self):
        q = 0.0130
        u = -0.027
        v = 0.021
        zero = 60

        psi = np.arange(0, 360, 22.5)
        flux_o, flux_e = get_flux_oe(1e5, psi, k=1.0, q=q, u=u, v=v, zero=zero)
        pol = SLSDualBeamPolarimetry(retarder='quarterwave', zero=60,
                                     compute_k=True)
        p = pol.compute(psi, flux_o, flux_e)

        assert_almost_equal(p.q.nominal, q)
        assert_almost_equal(p.u.nominal, u)
        assert_almost_equal(p.v.nominal, v)
        assert_almost_equal(p.k.nominal, 1.0, decimal=2)
        assert_almost_equal(p.zero, zero)

    @pytest.mark.parametrize('k', [0.99, 1.01, 1.03, 0.97])
    def test_fit_quarter_estimate_k(self, k):
        q = 0.130
        u = -0.027
        v = 0.021
        zero = 60

        psi = np.arange(0, 360, 22.5)
        flux_o, flux_e = get_flux_oe(1e5, psi, k=k, q=q, u=u, v=v, zero=zero)
        pol = SLSDualBeamPolarimetry(retarder='quarterwave', zero=60,
                                     compute_k=True)
        p = pol.compute(psi, flux_o, flux_e)

        assert_almost_equal(p.q.nominal, q, decimal=3)
        assert_almost_equal(p.u.nominal, u, decimal=3)
        assert_almost_equal(p.v.nominal, v, decimal=3)
        assert_almost_equal(p.k.nominal, k, decimal=2)
        assert_almost_equal(p.zero, zero, decimal=3)

    def test_fit_quarter_estimate_zero(self):
        q = 0.130
        u = -0.027
        v = 0.021
        zero = 60

        psi = np.arange(0, 360, 22.5)
        flux_o, flux_e = get_flux_oe(1e5, psi, k=1.0, q=q, u=u, v=v, zero=zero)
        pol = SLSDualBeamPolarimetry(retarder='quarterwave', compute_k=True)
        p = pol.compute(psi, flux_o, flux_e)

        assert_almost_equal(p.q.nominal, q, decimal=3)
        assert_almost_equal(p.u.nominal, u, decimal=3)
        assert_almost_equal(p.v.nominal, v, decimal=3)
        assert_almost_equal(p.k.nominal, 1.0, decimal=2)
        assert_almost_equal(p.zero.nominal, zero, decimal=3)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Module to handle online catalog queries."""

import numpy as np
from astroquery.simbad import Simbad
from astroquery.exceptions import TableParseError
from astropy.coordinates import SkyCoord, Angle
from ..astrometry.coords_utils import guess_coordinates
from ..py_utils import string_fix, process_list


MAX_PARALLEL_QUERY = 30
MAX_RETRIES_TIMEOUT = 10


def _timeout_retry(func, *args, **kwargs):
    """Retry a function until MAX_RETRIES_TIMEOUT is reached."""
    tried = kwargs.pop('_____retires', 0)
    try:
        q = func(*args, **kwargs)
    except (TimeoutError, TableParseError):
        if tried >= MAX_RETRIES_TIMEOUT:
            raise TimeoutError(f'TimeOut obtained in {MAX_RETRIES_TIMEOUT}'
                               ' tries, aborting.')
        return _timeout_retry(func, *args, **kwargs, _____retires=tried+1)
    return q


def _wrap_query_table(table):
    """Fix bytes and objects columns to strings."""
    for i in table.columns:
        tdtype = table[i].dtype.char
        if tdtype in ('b', 'B', 'S', 'a', 'O'):
            row = process_list(string_fix, table[i])
            table[i] = np.array(row, dtype=str)
    return table


def get_center_radius(ra, dec):
    """Get the center of a list RA and DEC.

    Parameters
    ----------
    ra, dec: pair of coordinate lists in decimal degrees.

    Returns
    -------
    center_ra: the RA of the center of the field.
    center_dec: the DEC of the center of the field.
    radius: The minimum radius to fit all objects in.
    """
    center_ra = (np.max(ra) + np.min(ra))/2
    center_dec = (np.max(dec) + np.min(dec))/2
    radius = np.max([np.max(ra) - np.min(ra),
                     np.max(dec) - np.min(dec)])
    return center_ra, center_dec, radius


def astroquery_skycoord(center, simbad=None):
    """Translate center coordinates to SkyCoord object.

    Notes
    -----
    - The supported formats are:
        - `str` or `bytes`: can be an object  name or
          `~astropy.coordinates.SkyCoord` compatible RA/DEC string.
        - `list`, `tuple` or `~numpy.array`: pair (RA, DEC) coordinates.
          Must be `~astropy.coordinates.SkyCoord` compatible.
        - `~astropy.coordinates.SkyCoord` object it self.
    """
    if isinstance(center, (str, bytes)):
        try:
            return SkyCoord(center)
        except (ValueError):
            if simbad is None:
                simbad = Simbad()
            t = simbad.query_object(center)
            if t is None:
                raise ValueError(f'Coordinates {center} could not be'
                                 ' resolved.')
            return guess_coordinates(t['RA'][0], t['DEC'][0], skycoord=True)
    elif isinstance(center, (tuple, list, np.ndarray)) and len(center) == 2:
        return guess_coordinates(center[0], center[1], skycoord=True)
    elif isinstance(center, SkyCoord):
        return center

    raise ValueError(f'Center coordinates {center} not undertood.')


def astroquery_radius(radius):
    """Convert several types of values to decimal degree angle radius.

    Notes
    -----
    - Current supported types are:
      - Any instance that can be converted to `float`.
      - `str` or `bytes` that can be converted to `astropy.coordinates.Angle`.
    """
    if isinstance(radius, (str, bytes)):
        radius = Angle(radius)

    if isinstance(radius, Angle):
        radius = radius.degree

    return f"{float(radius)}d"



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# flake8: noqa: F403, F405

import pytest
import numpy as np
from astropop.framedata import FrameData
from astropop.math.physical import QFloat, qfloat, units, \
                                   same_unit, UnitsError, \
                                   equal_within_errors, \
                                   convert_to_qfloat

from astropop.testing import *

# pylint: disable=no-member, pointless-statement


class Test_QFloat_UnitsHandling:
    def test_qfloat_same_unit(self):
        qf1 = QFloat(1.0, 0.1, 'm')
        qf2 = QFloat(200, 10, 'cm')
        qf3 = QFloat(120, 0.3, 's')
        qf4 = QFloat(2, 0.01, 'min')

        # must be ok, without any convertion
        qf_1, qf_2 = same_unit(qf1, qf1)
        assert_equal(qf_1.nominal, 1.0)
        assert_equal(qf_1.uncertainty, 0.1)
        assert_equal(qf_1.unit, units.m)
        assert_equal(qf_2.nominal, 1.0)
        assert_equal(qf_2.uncertainty, 0.1)
        assert_equal(qf_2.unit, units.m)

        # Doing conversion
        qf_1, qf_2 = same_unit(qf1, qf2)
        assert_equal(qf_1.unit, units.m)
        assert_equal(qf_1.nominal, 1.0)
        assert_equal(qf_1.uncertainty, 0.1)
        assert_equal(qf_2.unit, units.m)
        assert_equal(qf_2.nominal, 2.0)
        assert_equal(qf_2.uncertainty, 0.1)

        # Inverse Conversion
        qf_2, qf_1 = same_unit(qf2, qf1)
        assert_equal(qf_1.unit, units.cm)
        assert_equal(qf_1.nominal, 100)
        assert_equal(qf_1.uncertainty, 10)
        assert_equal(qf_2.unit, units.cm)
        assert_equal(qf_2.nominal, 200)
        assert_equal(qf_2.uncertainty, 10)

        # Raise conversion
        with pytest.raises(UnitsError):
            same_unit(qf3, qf1, self.test_qfloat_same_unit)

        # non-prefix conversion
        qf_3, qf_4 = same_unit(qf3, qf4)
        assert_equal(qf_3.nominal, 120)
        assert_equal(qf_3.uncertainty, 0.3)
        assert_equal(qf_3.unit, units.s)
        assert_equal(qf_4.nominal, 120)
        assert_equal(qf_4.uncertainty, 0.6)
        assert_equal(qf_4.unit, units.s)

        # Inverse non-prefix conversion
        qf_4, qf_3 = same_unit(qf4, qf3)
        assert_equal(qf_3.nominal, 2)
        assert_equal(qf_3.uncertainty, 0.005)
        assert_equal(qf_3.unit, units.minute)
        assert_equal(qf_4.nominal, 2)
        assert_equal(qf_4.uncertainty, 0.01)
        assert_equal(qf_4.unit, units.minute)

    def test_qfloat_same_unit_array(self):
        # like for single values, everything must run as arrays
        qf1 = QFloat(1.0, 0.1, 'm')
        qf2 = QFloat(np.arange(0, 2000, 100), np.arange(0, 100, 5), 'cm')
        qf3 = QFloat([120, 240], [0.3, 0.6], 's')
        qf4 = QFloat([2, 10], [0.01, 0.02], 'min')

        # Simplre conversion
        qf_1, qf_2 = same_unit(qf1, qf2, self.test_qfloat_same_unit_array)
        assert_equal(qf_1.nominal, 1.0)
        assert_equal(qf_1.uncertainty, 0.1)
        assert_equal(qf_1.unit, units.m)
        assert_almost_equal(qf_2.nominal, np.arange(0, 20, 1))
        assert_almost_equal(qf_2.std_dev, np.arange(0, 1, 0.05))
        assert_equal(qf_2.unit, units.m)

        # inverse conversion
        qf_2, qf_1 = same_unit(qf2, qf1, self.test_qfloat_same_unit_array)
        assert_equal(qf_1.nominal, 100)
        assert_equal(qf_1.uncertainty, 10)
        assert_equal(qf_1.unit, units.cm)
        assert_almost_equal(qf_2.nominal, np.arange(0, 2000, 100))
        assert_almost_equal(qf_2.std_dev, np.arange(0, 100, 5))
        assert_equal(qf_2.unit, units.cm)

        # incompatible
        with pytest.raises(UnitsError):
            same_unit(qf3, qf1, self.test_qfloat_same_unit)

        qf_3, qf_4 = same_unit(qf3, qf4, self.test_qfloat_same_unit_array)
        assert_almost_equal(qf_3.nominal, [120, 240])
        assert_almost_equal(qf_3.std_dev, [0.3, 0.6])
        assert_equal(qf_3.unit, units.s)
        assert_almost_equal(qf_4.nominal, [120, 600])
        assert_almost_equal(qf_4.std_dev, [0.6, 1.2])
        assert_equal(qf_4.unit, units.s)


class Test_QFloat_InitAndSet:
    @pytest.mark.parametrize('value, uncertainty, unit', [(1.0, 0.1, 'm'),
                                                          (200, None, None),
                                                          (120, 0.3, None),
                                                          (2.0, None, 'm'),
                                                          (np.arange(10), np.arange(10)*0.1, 'm'),
                                                          ([1, 2, 3], [0.1, 0.2, 0.3], None),
                                                          ([1, 2, 3], [0.1, 0.2, 0.3], 'm')])
    def test_qfloat_init_simple(self, value, uncertainty, unit):
        qf = QFloat(value, uncertainty, unit)
        assert_equal(qf.nominal, value)
        if uncertainty is None:
            assert_equal(qf.uncertainty, np.zeros_like(value))
        else:
            assert_equal(qf.uncertainty, uncertainty)
        if unit is None:
            assert_equal(qf.unit, units.dimensionless_unscaled)
        else:
            assert_equal(qf.unit, unit)

    @pytest.mark.parametrize('value', ['test', complex(1, 2), None, b'0\x00',
                                       [0, 'test', 2], [0, None, 0]])
    def test_qfloat_init_notreal(self, value):
        with pytest.raises(TypeError):
            QFloat(value)

    def test_qfloat_init_notreal_uncert(self):
        with pytest.raises(TypeError):
            QFloat(1, 'test')

        with pytest.raises(TypeError):
            QFloat(1, complex(1, 2))

        with pytest.raises(TypeError):
            QFloat([1, 2, 3], [1, 'Test', 2])

        qf = QFloat([1, 2, 3], [1, None, 3])
        assert_equal(qf.nominal, [1, 2, 3])
        assert_equal(qf.uncertainty, [1, 0, 3])

        with pytest.raises(TypeError):
            qf.uncertainty = [1, 'Test', 3]

    def test_qfloat_init_qfloat(self):
        qf1 = QFloat(1.0, 0.1, 'm')
        qf2 = QFloat(qf1)
        assert_equal(qf2.nominal, 1.0)
        assert_equal(qf2.uncertainty, 0.1)
        assert_equal(qf2.unit, units.m)
        assert_is_not(qf2, qf1)

    def test_qfloat_init_qfloat_conflicts(self):
        qf = QFloat(1.0, 0.1, 'm')
        with pytest.raises(ValueError):
            QFloat(qf, unit='m')
        with pytest.raises(ValueError):
            QFloat(qf, uncertainty=0.1)


class Test_QFloat_Repr:
    def test_qfloat_repr(self):
        assert_equal(repr(QFloat(1.0, 0.1, 'm')), '<QFloat 1.0+-0.1 m>')
        assert_equal(repr(QFloat(1, 0.1, 'm')), '<QFloat 1.0+-0.1 m>')
        assert_equal(repr(QFloat(1, 0.01, 'm')), '<QFloat 1.00+-0.01 m>')
        assert_equal(repr(QFloat(1.00, 0.01, 'm')), '<QFloat 1.00+-0.01 m>')
        assert_equal(repr(QFloat(1.00, 0.012, 'm')), '<QFloat 1.00+-0.01 m>')
        assert_equal(repr(QFloat(1.00, 0.016, 'm')), '<QFloat 1.00+-0.02 m>')
        assert_equal(repr(QFloat(1.002, 0.016, 'm')), '<QFloat 1.00+-0.02 m>')
        assert_equal(repr(QFloat(1.006, 0.016, 'm')), '<QFloat 1.01+-0.02 m>')
        assert_equal(repr(QFloat(1.000005, 0.01, 'm')), '<QFloat 1.00+-0.01 m>')
        assert_equal(repr(QFloat(9999, 1, 'm')), '<QFloat 9999+-1 m>')
        assert_equal(repr(QFloat(9999, 10, 'm')), '<QFloat 10000+-10 m>')

    def test_qfloat_repr_array(self):
        n = np.array([1, 1, 1, 1.01, 1.001, 1.006])
        s = np.array([0.1, 0.01, 1.0, 0.01, 0.01, 0.01])
        assert_equal(repr(QFloat(n, s, 'm')),
                     '<QFloat\narray([1.0+-0.1, 1.00+-0.01, 1+-1, 1.01+-0.01,'
                     ' 1.00+-0.01, 1.01+-0.01],\n      dtype=object)\nunit=m>')

        qf = QFloat([1, 2, 3, 4, 5], [0.1, 0.2, 0.3, 0.4, 0.5], 'm')
        assert_equal(repr(qf),
                     '<QFloat\narray([1.0+-0.1, 2.0+-0.2, 3.0+-0.3, 4.0+-0.4, 5.0'
                     '+-0.5], dtype=object)\nunit=m>')

        # very large array
        n = np.arange(0, 10000).reshape((100, 100))
        s = n*0.01
        qf = QFloat(n, s, 's')
        assert_equal(repr(qf),
                     '<QFloat\narray([[0+-0.0, 1.00+-0.01, 2.00+-0.02, ..., 97+-1'
                     ', 98+-1, 99+-1],\n       [100+-1, 101+-1, 102+-1, ..., 197+'
                     '-2, 198+-2, 199+-2],\n       [200+-2, 201+-2, 202+-2, ..., '
                     '297+-3, 298+-3, 299+-3],\n       ...,\n       [9700+-100, 9'
                     '700+-100, 9700+-100, ..., 9800+-100, 9800+-100,\n        98'
                     '00+-100],\n       [9800+-100, 9800+-100, 9800+-100, ..., 99'
                     '00+-100, 9900+-100,\n        9900+-100],\n       [9900+-100'
                     ', 9900+-100, 9900+-100, ..., 10000+-100, 10000+-100,\n     '
                     '   10000+-100]], dtype=object)\nunit=s>')


class Test_QFloat_Operators:
    def test_qfloat_properties_getset(self):
        # access all important properties
        qf = QFloat(5.0, 0.025, 'm')
        assert_equal(qf.uncertainty, 0.025)
        qf.uncertainty = 0.10
        assert_equal(qf.uncertainty, 0.10)
        assert_equal(qf.std_dev, 0.10)
        qf.uncertainty = 0.05
        assert_equal(qf.std_dev, 0.05)

        # setting nominal resets the uncertainty
        assert_equal(qf.nominal, 5.0)

        with pytest.raises(TypeError):
            qf.nominal = None

        assert_equal(qf.nominal, 5.0)
        qf.nominal = 10.0
        assert_equal(qf.nominal, 10.0)
        assert_equal(qf.std_dev, 0.0)
        assert_equal(qf.uncertainty, 0.0)

        # with arrays
        qf = QFloat([1, 2, 3], [0.1, 0.2, 0.3], 'm')
        assert_almost_equal(qf.uncertainty, [0.1, 0.2, 0.3])
        qf.uncertainty = [0.4, 0.5, 0.6]
        assert_almost_equal(qf.uncertainty, [0.4, 0.5, 0.6])
        assert_almost_equal(qf.std_dev, [0.4, 0.5, 0.6])
        qf.std_dev = [0.1, 0.2, 0.3]
        assert_almost_equal(qf.std_dev, [0.1, 0.2, 0.3])
        assert_almost_equal(qf.nominal, [1, 2, 3])

        with pytest.raises(TypeError):
            qf.nominal = None

        assert_almost_equal(qf.nominal, [1, 2, 3])
        qf.nominal = [4, 5, 6]
        assert_almost_equal(qf.nominal, [4, 5, 6])
        assert_almost_equal(qf.std_dev, [0, 0, 0])
        assert_almost_equal(qf.uncertainty, [0, 0, 0])

        # test value readonly property
        assert_equal(qf.value, [4, 5, 6])
        with pytest.raises(AttributeError):
            qf.value = 0

    def test_qfloat_set_nominal(self):
        qf = QFloat(5.0, 0.025, 'm')
        qf.nominal = 10.0
        assert_equal(qf.nominal, 10.0)
        assert_equal(qf.std_dev, 0.0)
        assert_equal(qf.uncertainty, 0.0)

        # with arrays
        qf = QFloat([1, 2, 3], [0.1, 0.2, 0.3], 'm')
        qf.nominal = [4, 5, 6]
        assert_almost_equal(qf.nominal, [4, 5, 6])
        assert_almost_equal(qf.std_dev, [0, 0, 0])
        assert_almost_equal(qf.uncertainty, [0, 0, 0])
        assert_equal(qf.unit, units.m)

        # tuple for (nominal, uncertainty)
        qf = QFloat(5.0, 0.025, 'm')
        qf.nominal = (10.0, 0.1)
        assert_equal(qf.nominal, 10.0)
        assert_equal(qf.uncertainty, 0.1)
        assert_equal(qf.unit, units.m)

        # tuple for (nominal, uncertainty, unit)
        qf = QFloat(5.0, 0.025, 'm')
        qf.nominal = (10.0, 0.1, 's')
        assert_equal(qf.nominal, 10.0)
        assert_equal(qf.uncertainty, 0.1)
        assert_equal(qf.unit, units.s)

        # raise non real
        with pytest.raises(TypeError):
            qf.nominal = None
        with pytest.raises(TypeError):
            qf.nominal = [1, 'test', 2]

    def test_qfloat_set_uncertainty(self):
        qf = QFloat(5.0, 0.025, 'm')
        qf.uncertainty = 0.1
        assert_equal(qf.uncertainty, 0.1)
        assert_equal(qf.std_dev, 0.1)

        # with arrays
        qf = QFloat([1, 2, 3], [0.1, 0.2, 0.3], 'm')
        qf.uncertainty = [0.4, 0.5, 0.6]
        assert_almost_equal(qf.uncertainty, [0.4, 0.5, 0.6])
        assert_almost_equal(qf.std_dev, [0.4, 0.5, 0.6])
        qf.uncertainty = None
        assert_almost_equal(qf.uncertainty, [0, 0, 0])

        # raise non real
        with pytest.raises(TypeError):
            qf.uncertainty = 'test'
        with pytest.raises(TypeError):
            qf.uncertainty = [1, 'test', 2]

        # raise invalid shape
        with pytest.raises(ValueError):
            qf.uncertainty = [1, 2, 3, 4]
        with pytest.raises(ValueError):
            qf.uncertainty = 1

    def test_qfloat_properties_reset(self):
        qf = QFloat(5.0, 0.025, 'm')
        i = id(qf)
        qf.reset(12, 0.2, 's')
        assert_equal(i, id(qf))
        assert_equal(qf, QFloat(12, 0.2, 's'))

        qf.reset([1, 2, 3], [0.1, 0.2, 0.3], 'm')
        assert_equal(i, id(qf))
        assert_equal(qf, QFloat([1, 2, 3], [0.1, 0.2, 0.3], 'm'))

    def test_qfloat_creation(self):
        def _create(*args, **kwargs):
            assert_is_instance(qfloat(*args, **kwargs), QFloat)

        def _raises(*args, **kwargs):
            with pytest.raises(Exception):
                qfloat(*args, **kwargs)

        # Simple scalar
        _create(1.0)
        _create(1.0, 0.1)
        _create(1.0, 0.1, 'm')

        # Arrays
        _create(np.ones(10))
        _create(np.ones(10), np.ones(10))
        _create(np.ones(10), np.ones(10), 'm')

        # Fails
        _raises(np.ones(10), 0.1)
        _raises(10.0, np.ones(10))
        _raises(uncertainty=0.1)
        _raises(unit='m')
        _raises(uncertainty=0.1, unit='m')

    def test_qfloat_always_positive_uncertainty(self):
        qf1 = QFloat(1.0, 0.1, 'm')
        assert_equal(qf1.uncertainty, 0.1)

        qf2 = QFloat(1.0, -0.1, 'm')
        assert_equal(qf2.uncertainty, 0.1)

        qf3 = QFloat(np.ones(10), np.ones(10)*0.1, 'm')
        assert_equal(qf3.uncertainty, np.ones(10)*0.1)

        qf4 = QFloat(np.ones(10), -np.ones(10)*0.1, 'm')
        assert_equal(qf4.uncertainty, np.ones(10)*0.1)

    @pytest.mark.parametrize('value,expect', [(QFloat(1.0, 0.1, 'm'), QFloat(1.0, 0.1, 'm')),
                                              (1, QFloat(1.0, 0, None)),
                                              (np.array([1, 2, 3]), QFloat([1, 2, 3], unit=None)),
                                              ('string', 'raise'),
                                              (None, 'raise'),
                                              (UnitsError, 'raise'),
                                              (1.0*units.m, QFloat(1.0, 0.0, 'm')),
                                              (FrameData(1.0, 'm', 'f8', 0.1), QFloat(1.0, 0.1, 'm'))])
    def test_qfloat_converttoqfloat(self, value, expect):
        if expect == 'raise':
            with pytest.raises(Exception):
                convert_to_qfloat(value)
        else:
            conv = convert_to_qfloat(value)
            assert_equal(conv.nominal, expect.nominal)
            assert_equal(conv.uncertainty, expect.uncertainty)
            assert_equal(conv.unit, expect.unit)

    def test_qfloat_unit_conversion(self):
        qf1 = QFloat(1.0, 0.01, 'm')

        # test converting with string
        qf2 = qf1 << 'cm'
        assert_equal(qf2.nominal, 100)
        assert_equal(qf2.uncertainty, 1)
        assert_equal(qf2.unit, units.cm)

        # test converting using instance
        qf3 = qf1 << units.cm
        assert_equal(qf3.nominal, 100)
        assert_equal(qf3.uncertainty, 1)
        assert_equal(qf3.unit, units.cm)

        # But qf1 must stay the same
        assert_equal(qf1.nominal, 1.0)
        assert_equal(qf1.uncertainty, 0.01)
        assert_equal(qf1.unit, units.m)

        with pytest.raises(Exception):
            # None must fail
            qf1 << None

        with pytest.raises(Exception):
            # incompatible must fail
            qf1 << units.s

        # and this must work with arrays
        qf1 = QFloat([1, 2], [0.1, 0.2], unit='m')
        qf2 = qf1 << 'km'
        assert_equal(qf2.nominal, [0.001, 0.002])
        assert_equal(qf2.uncertainty, [0.0001, 0.0002])
        assert_equal(qf2.unit, units.km)

    def test_qfloat_unit_conversion_to(self):
        qf1 = QFloat(1.0, 0.01, 'm')

        # test converting with string
        qf2 = qf1.to('cm')
        assert_equal(qf2.nominal, 100)
        assert_equal(qf2.uncertainty, 1)
        assert_equal(qf2.unit, units.cm)

        # test converting using instance
        qf3 = qf1.to(units.cm)
        assert_equal(qf3.nominal, 100)
        assert_equal(qf3.uncertainty, 1)
        assert_equal(qf3.unit, units.cm)

        # But qf1 must stay the same
        assert_equal(qf1.nominal, 1.0)
        assert_equal(qf1.uncertainty, 0.01)
        assert_equal(qf1.unit, units.m)

        # inline: now qf1 must change
        i = id(qf1)
        qf1 <<= 'cm'
        assert_equal(qf1.nominal, 100)
        assert_equal(qf1.uncertainty, 1)
        assert_equal(qf1.unit, units.cm)
        assert_equal(id(qf1), i)

        with pytest.raises(Exception):
            # None must fail
            qf1.to(None)

        with pytest.raises(Exception):
            # incompatible must fail
            qf1.to(units.s)

        # and this must work with arrays
        qf1 = QFloat([1, 2], [0.1, 0.2], unit='m')
        qf2 = qf1.to('km')
        assert_equal(qf2.nominal, [0.001, 0.002])
        assert_equal(qf2.uncertainty, [0.0001, 0.0002])
        assert_equal(qf2.unit, units.km)

    def test_qfloat_getitem(self):
        # simple array
        qf = QFloat([1, 2, 3, 4, 5], [0.1, 0.2, 0.3, 0.4, 0.5], 's')

        qf1 = qf[0]
        assert_equal(qf1.nominal, 1)
        assert_equal(qf1.uncertainty, 0.1)
        assert_equal(qf1.unit, units.s)

        qf3 = qf[2]
        assert_equal(qf3.nominal, 3)
        assert_equal(qf3.uncertainty, 0.3)
        assert_equal(qf3.unit, units.s)

        qf4 = qf[-1]
        assert_equal(qf4.nominal, 5)
        assert_equal(qf4.uncertainty, 0.5)
        assert_equal(qf4.unit, units.s)

        qf5 = qf[1:4]
        assert_equal(qf5.nominal, [2, 3, 4])
        assert_equal(qf5.uncertainty, [0.2, 0.3, 0.4])
        assert_equal(qf5.unit, units.s)

        with pytest.raises(IndexError):
            qf[10]

        # 2D array
        qf = QFloat(np.arange(1, 17, 1).reshape((4, 4)),
                    np.arange(1, 17, 1).reshape((4, 4))*0.01, 'm')

        qfrow = qf[0]
        assert_equal(qfrow.nominal, [1, 2, 3, 4])
        assert_equal(qfrow.uncertainty, [0.01, 0.02, 0.03, 0.04])
        assert_equal(qfrow.unit, units.m)

        qfcol = qf[:, 1]
        assert_equal(qfcol.nominal, [2, 6, 10, 14])
        assert_equal(qfcol.uncertainty, [0.02, 0.06, 0.1, 0.14])
        assert_equal(qfcol.unit, units.m)

        qf0 = qf[0, 0]
        assert_equal(qf0.nominal, 1)
        assert_equal(qf0.uncertainty, 0.01)
        assert_equal(qf0.unit, units.m)

        qf1 = qf[-1, -1]
        assert_equal(qf1.nominal, 16)
        assert_equal(qf1.uncertainty, 0.16)
        assert_equal(qf1.unit, units.m)

        qfs = qf[2:, 1:3]
        assert_equal(qfs.nominal, [[10, 11], [14, 15]])
        assert_equal(qfs.uncertainty, [[0.10, 0.11], [0.14, 0.15]])
        assert_equal(qfs.unit, units.m)

        with pytest.raises(IndexError):
            qf[10]

        with pytest.raises(IndexError):
            qf[0, 10]

        # Not iterable
        qf = QFloat(10, 0.1, 'm')
        with pytest.raises(TypeError):
            qf[0]

    def test_qfloat_setitem(self):
        # simple array
        qf = QFloat([1, 2, 3, 4, 5], [0.1, 0.2, 0.3, 0.4, 0.5], 's')
        qf[0] = QFloat(10, 0.5, 's')
        qf[-1] = QFloat(-10, unit='min')
        qf[2:4] = QFloat(1, 0.3, 's')
        assert_equal(qf.nominal, [10, 2, 1, 1, -600])
        assert_equal(qf.uncertainty, [0.5, 0.2, 0.3, 0.3, 0])
        assert_equal(qf.unit, units.s)

        with pytest.raises(IndexError):
            qf[5] = QFloat(10, 0.1, 's')
        with pytest.raises(UnitsError):
            qf[0] = 10
        with pytest.raises(UnitsError):
            qf[0] = QFloat(10, 0.1, 'm')

        # 2D array
        qf = QFloat(np.arange(1, 17, 1).reshape((4, 4)),
                    np.arange(1, 17, 1).reshape((4, 4))*0.01,
                    'm')
        qf[0] = QFloat(np.arange(5, 9, 1), np.arange(0.3, 0.7, 0.1), 'm')
        qf[:, 0] = QFloat(np.arange(1, 5, 1), [0.1, 0.3, 0.9, 0.4], 'm')
        qf[2, 2] = QFloat(1000, 10, 'cm')
        qf[2:, 2:] = QFloat(20, 0.66, 'm')
        assert_equal(qf.nominal, [[1, 6, 7, 8], [2, 6, 7, 8], [3, 10, 20, 20], [4, 14, 20, 20]])
        assert_almost_equal(qf.uncertainty, [[0.1, 0.4, 0.5, 0.6],
                                             [0.3, 0.06, 0.07, 0.08],
                                             [0.9, 0.1, 0.66, 0.66],
                                             [0.4, 0.14, 0.66, 0.66]])
        assert_equal(qf.unit, units.m)

        with pytest.raises(IndexError):
            qf[5, 10] = QFloat(10, 0.1, 'm')
        with pytest.raises(UnitsError):
            qf[0, 0] = 10
        with pytest.raises(UnitsError):
            qf[0, 0] = QFloat(10, 0.1, 's')

    def test_qfloat_len(self):
        with pytest.raises(TypeError):
            len(QFloat(1.0, 0.1, 'm'))
        assert_equal(len(QFloat([1], [0.1], 'm')), 1)
        assert_equal(len(QFloat([2, 3], [1, 2], 'm')), 2)

        # same behavior of numpy
        assert_equal(len(QFloat(np.zeros((10, 5)), np.zeros((10, 5)), 'm')), 10)
        assert_equal(len(QFloat(np.zeros((10, 5)), np.zeros((10, 5)), 'm')[0]), 5)


class Test_QFloat_Comparison:
    def test_qfloat_comparison_equality_same_unit(self):
        # These numbers must not be equal, but equal within errors.
        qf1 = QFloat(1.0, 0.1, 'm')
        qf2 = QFloat(1.05, 0.2, 'm')
        qf3 = QFloat(1.8, 0.1, 'm')
        assert_false(qf1 == qf2)
        assert_true(qf1 != qf2)
        assert_true(qf1 != qf3)
        assert_true(qf2 != qf3)
        assert_true(equal_within_errors(qf1, qf2))
        assert_false(equal_within_errors(qf1, qf3))

    def test_qfloat_comparison_equality_convert_units(self):
        # Units must matter
        qf1 = QFloat(1.0, 0.1, 'm')
        qf3 = QFloat(1.0, 0.1, 'cm')
        qf4 = QFloat(100, 10, 'cm')
        assert_false(qf1 == qf3)
        assert_false(qf1 != qf4)
        assert_true(qf1 != qf3)
        assert_true(qf1 == qf4)
        assert_false(equal_within_errors(qf1, qf3))
        assert_true(equal_within_errors(qf1, qf4))

    def test_qfloat_comparison_equality_incompatible_units(self):
        qf1 = QFloat(1.0, 0.1, 'm')
        qf2 = QFloat(1.0, 0.1, 's')
        assert_false(qf1 == qf2)
        assert_true(qf1 != qf2)
        assert_false(equal_within_errors(qf1, qf2))

    def test_qfloat_comparison_inequality_same_unit(self):
        qf1 = QFloat(1.0, 0.1, 'm')
        qf2 = QFloat(1.0, 0.2, 'm')
        qf3 = QFloat(1.1, 0.1, 'm')
        qf4 = QFloat(1.1, 0.2, 'm')
        qf5 = QFloat(0.9, 0.1, 'm')
        qf6 = QFloat(0.9, 0.5, 'm')

        assert_true(qf1 <= qf2)
        assert_false(qf1 < qf2)
        assert_true(qf1 >= qf2)
        assert_false(qf1 > qf2)

        assert_true(qf1 <= qf3)
        assert_true(qf1 < qf3)
        assert_false(qf1 > qf3)
        assert_false(qf1 >= qf3)

        assert_true(qf1 <= qf4)
        assert_true(qf1 < qf4)
        assert_false(qf1 > qf4)
        assert_false(qf1 >= qf4)

        assert_true(qf1 >= qf5)
        assert_true(qf1 > qf5)
        assert_false(qf1 < qf5)
        assert_false(qf1 <= qf5)

        assert_true(qf1 >= qf6)
        assert_true(qf1 > qf6)
        assert_false(qf1 < qf6)
        assert_false(qf1 <= qf6)

    def test_qfloat_comparison_inequality_convert_unit(self):
        qf1 = QFloat(1.0, 0.1, 'm')
        qf2 = QFloat(200, 0.1, 'cm')
        qf3 = QFloat(0.005, 0.0003, 'km')
        qf4 = QFloat(0.0001, 0.00001, 'km')

        assert_true(qf2 > qf1)
        assert_true(qf2 >= qf1)
        assert_false(qf2 < qf1)
        assert_false(qf2 <= qf1)

        assert_true(qf3 > qf1)
        assert_true(qf3 >= qf1)
        assert_false(qf3 < qf1)
        assert_false(qf3 <= qf1)

        assert_true(qf4 < qf1)
        assert_true(qf4 <= qf1)
        assert_false(qf4 > qf1)
        assert_false(qf4 >= qf1)

    def test_qfloat_comparison_inequality_incompatible_units(self):
        qf1 = QFloat(1.0, 0.1, 'm')
        qf2 = QFloat(1.0, 0.1, 's')

        with pytest.raises(UnitsError):
            qf1 < qf2

        with pytest.raises(UnitsError):
            qf1 <= qf2

        with pytest.raises(UnitsError):
            qf1 > qf2

        with pytest.raises(UnitsError):
            qf1 >= qf2


class Test_QFloat_Add:
    def test_qfloat_math_add_single(self):
        qf1 = QFloat(1.0, 0.01, 'm')
        qf2 = QFloat(50, 2, 'cm')
        qf3 = QFloat(10, 0.1, None)
        qf4 = QFloat(60, 0.001, 's')

        res1 = qf1 + qf2
        assert_equal(res1.nominal, 1.5)
        assert_almost_equal(res1.uncertainty, 0.022360679774)
        assert_equal(res1.unit, units.m)

        # same as above, but in cm
        res2 = qf2 + qf1
        assert_equal(res2.nominal, 150)
        assert_almost_equal(res2.uncertainty, 2.2360679774)
        assert_equal(res2.unit, units.cm)

        # Should fail with incompatible units
        with pytest.raises(UnitsError):
            qf1 + qf3

        with pytest.raises(UnitsError):
            qf1 + qf4

    def test_qfloat_math_add_array(self):
        qf1 = QFloat([1, 2, 3, 4], [0.01, 0.02, 0.01, 0.02], 'm')
        qf2 = QFloat([150, 250, 50, 550], [0.1, 5, 0.4, 2], 'cm')
        qf3 = QFloat(1, 0.01, 'm')
        qf4 = QFloat(10, 0.1, None)
        qf5 = QFloat(60, 0.001, 's')

        res1 = qf1 + qf2
        assert_equal(res1.nominal, [2.5, 4.5, 3.5, 9.5])
        assert_almost_equal(res1.uncertainty, [0.01004987562112089, 0.05385164807134505,
                                               0.010770329614269008, 0.0282842712474619])
        assert_equal(res1.unit, units.m)

        # same as above, but in cm
        res2 = qf2 + qf1
        assert_equal(res2.nominal, [250, 450, 350, 950])
        assert_almost_equal(res2.uncertainty, [1.004987562112089, 5.385164807134505,
                                               1.0770329614269008, 2.82842712474619])
        assert_equal(res2.unit, units.cm)

        # Numpy behavior is to sum arrays with single numbers
        res3 = qf1 + qf3
        assert_equal(res3.nominal, [2.0, 3.0, 4.0, 5.0])
        assert_almost_equal(res3.uncertainty, [0.01414213562373095, 0.022360679774997897,
                                               0.01414213562373095, 0.022360679774997897])
        assert_equal(res3.unit, units.m)

        # So, it should sum numbers with arrays
        res4 = qf3 + qf1
        assert_equal(res4.nominal, [2.0, 3.0, 4.0, 5.0])
        assert_almost_equal(res4.uncertainty, [0.01414213562373095, 0.022360679774997897,
                                               0.01414213562373095, 0.022360679774997897])
        assert_equal(res4.unit, units.m)

        # Should fail with incompatible units
        with pytest.raises(UnitsError):
            qf1 + qf4

        with pytest.raises(UnitsError):
            qf1 + qf5

    def test_qfloat_math_add_with_numbers(self):
        qf1 = QFloat(1, 0.1, 'm')
        qf2 = QFloat(1, 0.1)  # dimensionless
        qf3 = QFloat([2, 3], [0.1, 0.2])
        qf4 = QFloat([2, 3], [0.1, 0.2], 's')

        # Should raise with dimension measurements
        with pytest.raises(UnitsError):
            qf1 + 1
        with pytest.raises(UnitsError):
            1 + qf1

        # but works with dimensionless
        res1 = qf2 + 2
        assert_equal(res1.nominal, 3)
        assert_almost_equal(res1.uncertainty, 0.1)
        assert_equal(res1.unit, units.dimensionless_unscaled)

        # same as above, but inverse
        res2 = 2 + qf2
        assert_equal(res2.nominal, 3)
        assert_almost_equal(res2.uncertainty, 0.1)
        assert_equal(res2.unit, units.dimensionless_unscaled)

        # and with arrays!
        res3 = qf3 + 1
        assert_equal(res3.nominal, [3, 4])
        assert_almost_equal(res3.uncertainty, [0.1, 0.2])
        assert_equal(res3.unit, units.dimensionless_unscaled)

        # and with arrays inverse!
        res4 = 1 + qf3
        assert_equal(res4.nominal, [3, 4])
        assert_almost_equal(res4.uncertainty, [0.1, 0.2])
        assert_equal(res4.unit, units.dimensionless_unscaled)

        # array array
        with pytest.raises(UnitsError):
            qf4 + [1, 2]

        with pytest.raises(UnitsError):
            [1, 2] + qf4

        res5 = qf3 + [1, 2]
        assert_equal(res5.nominal, [3, 5])
        assert_almost_equal(res5.uncertainty, [0.1, 0.2])
        assert_equal(res5.unit, units.dimensionless_unscaled)

        res6 = [1, 2] + qf3
        assert_equal(res6.nominal, [3, 5])
        assert_almost_equal(res6.uncertainty, [0.1, 0.2])
        assert_equal(res6.unit, units.dimensionless_unscaled)

    def test_qfloat_math_add_inline(self):
        # Sum with another qfloat
        qf = QFloat(1.0, 0.1, 'm')
        i = id(qf)
        qf += QFloat(2.0, 0.1, 'm')
        assert_equal(qf.nominal, 3)
        assert_almost_equal(qf.uncertainty, 0.1414213562373095)
        assert_equal(qf.unit, units.m)
        assert_equal(id(qf), i)

        # raise if incompatible
        with pytest.raises(UnitsError):
            qf += QFloat(3, unit='s')
        assert_equal(id(qf), i)

        with pytest.raises(UnitsError):
            qf += 1
        assert_equal(id(qf), i)

        # Arrays
        qf = QFloat([1, 2, 3, 4], [0.01, 0.02, 0.03, 0.04], 'cm')
        i = id(qf)
        qf += QFloat(1, 0.01, 'cm')
        assert_equal(qf.nominal, [2, 3, 4, 5])
        assert_almost_equal(qf.uncertainty, [0.01414213562373095, 0.022360679774997897,
                                             0.03162277660168379, 0.04123105625617661])
        assert_equal(qf.unit, units.cm)
        assert_equal(id(qf), i)

        with pytest.raises(UnitsError):
            qf += [2, 3, 4, 5]
        assert_equal(id(qf), i)


class Test_QFloat_Sub:
    def test_qfloat_math_sub_single(self):
        qf1 = QFloat(1.0, 0.01, 'm')
        qf2 = QFloat(50, 2, 'cm')
        qf3 = QFloat(10, 0.1, None)
        qf4 = QFloat(60, 0.001, 's')

        res1 = qf1 - qf2
        assert_equal(res1.nominal, 0.5)
        assert_almost_equal(res1.uncertainty, 0.022360679774)
        assert_equal(res1.unit, units.m)

        # same as above, but in cm
        res2 = qf2 - qf1
        assert_equal(res2.nominal, -50)
        assert_almost_equal(res2.uncertainty, 2.2360679774)
        assert_equal(res2.unit, units.cm)

        # Should fail with incompatible units
        with pytest.raises(UnitsError):
            qf1 - qf3

        with pytest.raises(UnitsError):
            qf1 - qf4

    def test_qfloat_math_sub_array(self):
        qf1 = QFloat([1, 2, 3, 4], [0.01, 0.02, 0.01, 0.02], 'm')
        qf2 = QFloat([50, 40, 30, 550], [0.1, 5, 0.4, 2], 'cm')
        qf3 = QFloat(1, 0.01, 'm')
        qf4 = QFloat(10, 0.1, None)
        qf5 = QFloat(60, 0.001, 's')

        res1 = qf1 - qf2
        assert_equal(res1.nominal, [0.5, 1.6, 2.7, -1.5])
        assert_almost_equal(res1.uncertainty, [0.01004987562112089, 0.05385164807134505,
                                               0.010770329614269008, 0.0282842712474619])
        assert_equal(res1.unit, units.m)

        # same as above, but in cm
        res2 = qf2 - qf1
        assert_equal(res2.nominal, [-50, -160, -270, 150])
        assert_almost_equal(res2.uncertainty, [1.004987562112089, 5.385164807134505,
                                               1.0770329614269008, 2.82842712474619])
        assert_equal(res2.unit, units.cm)

        # Numpy behavior is to sum arrays with single numbers
        res3 = qf1 - qf3
        assert_equal(res3.nominal, [0.0, 1.0, 2.0, 3.0])
        assert_almost_equal(res3.uncertainty, [0.01414213562373095, 0.022360679774997897,
                                               0.01414213562373095, 0.022360679774997897])
        assert_equal(res3.unit, units.m)

        # So, it should sum numbers with arrays
        res4 = qf3 - qf1
        assert_equal(res4.nominal, [0.0, -1, -2.0, -3.0])
        assert_almost_equal(res4.uncertainty, [0.01414213562373095, 0.022360679774997897,
                                               0.01414213562373095, 0.022360679774997897])
        assert_equal(res4.unit, units.m)

        # Should fail with incompatible units
        with pytest.raises(UnitsError):
            qf1 - qf4

        with pytest.raises(UnitsError):
            qf1 - qf5

    def test_qfloat_math_sub_with_numbers(self):
        qf1 = QFloat(1, 0.1, 'm')
        qf2 = QFloat(1, 0.1)  # dimensionless
        qf3 = QFloat([2, 3], [0.1, 0.2])
        qf4 = QFloat([2, 3], [0.1, 0.2], 's')

        # Should raise with dimension measurements
        with pytest.raises(UnitsError):
            qf1 - 1
        with pytest.raises(UnitsError):
            1 - qf1

        # but works with dimensionless
        res1 = qf2 - 2
        assert_equal(res1.nominal, -1)
        assert_almost_equal(res1.uncertainty, 0.1)
        assert_equal(res1.unit, units.dimensionless_unscaled)

        # same as above, but inverse
        res2 = 2 - qf2
        assert_equal(res2.nominal, 1)
        assert_almost_equal(res2.uncertainty, 0.1)
        assert_equal(res2.unit, units.dimensionless_unscaled)

        # and with arrays!
        res3 = qf3 - 1
        assert_equal(res3.nominal, [1, 2])
        assert_almost_equal(res3.uncertainty, [0.1, 0.2])
        assert_equal(res3.unit, units.dimensionless_unscaled)

        # and with arrays inverse!
        res4 = 1 - qf3
        assert_equal(res4.nominal, [-1, -2])
        assert_almost_equal(res4.uncertainty, [0.1, 0.2])
        assert_equal(res4.unit, units.dimensionless_unscaled)

        # array array
        with pytest.raises(UnitsError):
            qf4 - [1, 2]

        with pytest.raises(UnitsError):
            [1, 2] - qf4

        res5 = qf3 - [1, 2]
        assert_equal(res5.nominal, [1, 1])
        assert_almost_equal(res5.uncertainty, [0.1, 0.2])
        assert_equal(res5.unit, units.dimensionless_unscaled)

        res6 = [1, 2] - qf3
        assert_equal(res6.nominal, [-1, -1])
        assert_almost_equal(res6.uncertainty, [0.1, 0.2])
        assert_equal(res6.unit, units.dimensionless_unscaled)

    def test_qfloat_math_sub_inline(self):
        # Sum with another qfloat
        qf = QFloat(1.0, 0.01, 'm')
        i = id(qf)
        qf -= QFloat(0.5, 0.01, 'm')
        assert_equal(qf.nominal, 0.5)
        assert_almost_equal(qf.uncertainty, 0.01414213562373095)
        assert_equal(qf.unit, units.m)
        assert_equal(id(qf), i)

        # raise if incompatible
        with pytest.raises(UnitsError):
            qf -= QFloat(3, unit='s')
        assert_equal(id(qf), i)

        with pytest.raises(UnitsError):
            qf += 1
        assert_equal(id(qf), i)

        # Arrays
        qf = QFloat([1, 2, 3, 4], [0.01, 0.02, 0.03, 0.04], 'cm')
        i = id(qf)
        qf -= QFloat(1, 0.01, 'cm')
        assert_equal(qf.nominal, [0, 1, 2, 3])
        assert_almost_equal(qf.uncertainty, [0.01414213562373095,
                                             0.022360679774997897,
                                             0.03162277660168379,
                                             0.04123105625617661])
        assert_equal(qf.unit, units.cm)
        assert_equal(id(qf), i)

        with pytest.raises(UnitsError):
            qf -= [2, 3, 4, 5]
        assert_equal(id(qf), i)


class Test_QFloat_Mul:
    def test_qfloat_math_mul_single(self):
        qf1 = QFloat(30, 0.5, 's')
        qf2 = QFloat(10, 0.1, 'm')
        qf3 = QFloat(200, 1, 'cm')
        qf4 = QFloat(5, 0.001, None)

        # different dimensions
        res1 = qf1 * qf2
        assert_equal(res1.nominal, 300)
        assert_equal(res1.uncertainty, 5.830951894845301)
        assert_equal(res1.unit, units.s*units.m)
        # inverse gets just the same
        res2 = qf2 * qf1
        assert_equal(res2.nominal, 300)
        assert_equal(res2.uncertainty, 5.830951894845301)
        assert_equal(res2.unit, units.s*units.m)

        # same dimension. Astropy behavior is not convert the units
        res3 = qf2 * qf3
        assert_equal(res3.nominal, 2000)
        assert_equal(res3.uncertainty, 22.360679774997898)
        assert_equal(res3.unit, units.cm*units.m)
        # inverse the same
        res4 = qf2 * qf3
        assert_equal(res4.nominal, 2000)
        assert_equal(res4.uncertainty, 22.360679774997898)
        assert_equal(res4.unit, units.cm*units.m)

        # None (dimensionless)
        res5 = qf1 * qf4
        assert_equal(res5.nominal, 150)
        assert_equal(res5.uncertainty, 2.5001799935204665)
        assert_equal(res5.unit, units.s)
        # inverse the same
        res5 = qf4 * qf1
        assert_equal(res5.nominal, 150)
        assert_equal(res5.uncertainty, 2.5001799935204665)
        assert_equal(res5.unit, units.s)

        # With numbers
        res6 = qf1 * 2
        assert_equal(res6.nominal, 60)
        assert_equal(res6.uncertainty, 1.0)
        assert_equal(res6.unit, units.s)
        # Inverse same
        res6 = 2 * qf1
        assert_equal(res6.nominal, 60)
        assert_equal(res6.uncertainty, 1.0)
        assert_equal(res6.unit, units.s)

        # With units
        res7 = qf1 * units.m
        assert_equal(res7.nominal, 30)
        assert_equal(res7.uncertainty, 0.5)
        assert_equal(res7.unit, units.m*units.s)
        # And with string!
        res9 = qf1 * 'm'
        assert_equal(res9.nominal, 30)
        assert_equal(res9.uncertainty, 0.5)
        assert_equal(res9.unit, units.m*units.s)
        # inverse
        res_a = 'm' * qf1
        assert_equal(res_a.nominal, 30)
        assert_equal(res_a.uncertainty, 0.5)
        assert_equal(res_a.unit, units.m*units.s)

    def test_qfloat_math_mul_array(self):
        qf1 = QFloat(np.arange(5), np.arange(5)*0.01, 'm')
        qf2 = QFloat(20, 0.1, 's')
        qf3 = QFloat(np.arange(16).reshape((4, 4)),
                     np.arange(16).reshape((4, 4))*0.0001, 'km')

        res1 = qf1 * qf1
        assert_equal(res1.nominal, [0, 1, 4, 9, 16])
        assert_almost_equal(res1.uncertainty, [0, 0.01414213562373095, 0.0565685424949238,
                                               0.12727922061357855, 0.2262741699796952])
        assert_equal(res1.unit, units.m*units.m)

        res2 = qf1 * qf2
        assert_equal(res2.nominal, [0, 20, 40, 60, 80])
        assert_almost_equal(res2.uncertainty, [0, 0.223606797749979, 0.447213595499958,
                                               0.6708203932499369, 0.894427190999916])
        assert_equal(res2.unit, units.m*units.s)
        # inverse same
        res3 = qf2 * qf1
        assert_equal(res3.nominal, [0, 20, 40, 60, 80])
        assert_almost_equal(res3.uncertainty, [0, 0.223606797749979, 0.447213595499958,
                                               0.6708203932499369, 0.894427190999916])
        assert_equal(res3.unit, units.m*units.s)

        # with numbers
        res4 = qf1 * 10
        assert_equal(res4.nominal, [0, 10, 20, 30, 40])
        assert_equal(res4.uncertainty, [0, 0.1, 0.2, 0.3, 0.4])
        assert_equal(res4.unit, units.m)
        # same inverse
        res5 = 10 * qf1
        assert_equal(res5.nominal, [0, 10, 20, 30, 40])
        assert_equal(res5.uncertainty, [0, 0.1, 0.2, 0.3, 0.4])
        assert_equal(res5.unit, units.m)

        # 2D Array with qfloat
        res6 = qf2 * qf3
        assert_equal(res6.nominal, [[0, 20, 40, 60],
                                    [80, 100, 120, 140],
                                    [160, 180, 200, 220],
                                    [240, 260, 280, 300]])
        assert_almost_equal(res6.uncertainty, [[0, 0.10002, 0.20004, 0.30005999],
                                               [0.40007999, 0.50009999, 0.60011999, 0.70013999],
                                               [0.80015998, 0.90017998, 1.00019998, 1.10021998],
                                               [1.20023998, 1.30025997, 1.40027997, 1.50029997]])
        assert_equal(res6.unit, units.km*units.s)
        # inverse same
        res7 = qf3 * qf2
        assert_equal(res7.nominal, [[0, 20, 40, 60],
                                    [80, 100, 120, 140],
                                    [160, 180, 200, 220],
                                    [240, 260, 280, 300]])
        assert_almost_equal(res7.uncertainty, [[0, 0.10002, 0.20004, 0.30005999],
                                               [0.40007999, 0.50009999, 0.60011999, 0.70013999],
                                               [0.80015998, 0.90017998, 1.00019998, 1.10021998],
                                               [1.20023998, 1.30025997, 1.40027997, 1.50029997]])
        assert_equal(res7.unit, units.km*units.s)

        # 2D array with numbers
        res8 = 3 * qf3
        assert_equal(res8.nominal,
                     np.arange(16).reshape((4, 4))*3)
        assert_almost_equal(res8.uncertainty,
                            np.arange(16).reshape((4, 4))*0.0001*3)
        assert_equal(res8.unit, units.km)
        # same inverse
        res9 = qf3 * 3
        assert_equal(res9.nominal,
                     np.arange(16).reshape((4, 4))*3)
        assert_almost_equal(res9.uncertainty,
                            np.arange(16).reshape((4, 4))*0.0001*3)
        assert_equal(res9.unit, units.km)

        # With units
        res_a = qf1 * units.m
        assert_equal(res_a.nominal, qf1.nominal)
        assert_equal(res_a.uncertainty, qf1.uncertainty)
        assert_equal(res_a.unit, units.m*units.m)
        # And with string!
        res_c = qf1 * 'm'
        assert_equal(res_c.nominal, qf1.nominal)
        assert_equal(res_c.uncertainty, qf1.uncertainty)
        assert_equal(res_c.unit, units.m*units.m)
        # inverse
        res_b = 'm' * qf1
        assert_equal(res_b.nominal, qf1.nominal)
        assert_equal(res_b.uncertainty, qf1.uncertainty)
        assert_equal(res_b.unit, units.m*units.m)

    def test_qfloat_math_mul_inline(self):
        # single number
        qf1 = QFloat(30, 0.5, 's')
        i = id(qf1)
        qf1 *= QFloat(5, 0.01, None)
        assert_equal(qf1.nominal, 150)
        assert_almost_equal(qf1.uncertainty, 2.5179356624028344)
        assert_equal(qf1.unit, units.s)
        assert_equal(i, id(qf1))
        qf1 *= QFloat(2, 0.1, 'm')
        assert_equal(qf1.nominal, 300)
        assert_almost_equal(qf1.uncertainty, 15.822768405054788)
        assert_equal(qf1.unit, units.s*units.m)
        assert_equal(i, id(qf1))
        qf1 *= 10
        assert_equal(qf1.nominal, 3000)
        assert_almost_equal(qf1.uncertainty, 158.22768405054788)
        assert_equal(qf1.unit, units.s*units.m)
        assert_equal(i, id(qf1))
        qf1 *= 'g'
        assert_equal(qf1.nominal, 3000)
        assert_almost_equal(qf1.uncertainty, 158.22768405054788)
        assert_equal(qf1.unit, units.s*units.m*units.g)
        assert_equal(i, id(qf1))
        qf1 *= units.K
        assert_equal(qf1.nominal, 3000)
        assert_almost_equal(qf1.uncertainty, 158.22768405054788)
        assert_equal(qf1.unit, units.s*units.m*units.g*units.K)
        assert_equal(i, id(qf1))

        # array
        qf2 = QFloat([2, 3, 4], [0.1, 0.2, 0.3], 'm')
        i = id(qf2)
        qf2 *= QFloat(5, 0.01, None)
        assert_almost_equal(qf2.nominal, [10, 15, 20])
        assert_almost_equal(qf2.uncertainty, [0.50039984, 1.0004499, 1.50053324])
        assert_equal(qf2.unit, units.m)
        assert_equal(i, id(qf2))
        qf2 *= QFloat(3.5, 0.1, 'm')
        assert_almost_equal(qf2.nominal, [35, 52.5, 70])
        assert_almost_equal(qf2.uncertainty, [2.01677961, 3.80933393, 5.61979537])
        assert_equal(qf2.unit, units.m*units.m)
        assert_equal(i, id(qf2))
        qf2 *= 10
        assert_almost_equal(qf2.nominal, [350, 525, 700])
        assert_almost_equal(qf2.uncertainty, [20.1677961, 38.0933393, 56.1979537])
        assert_equal(qf2.unit, units.m*units.m)
        assert_equal(i, id(qf2))
        qf2 *= 'g'
        assert_almost_equal(qf2.nominal, [350, 525, 700])
        assert_almost_equal(qf2.uncertainty, [20.1677961, 38.0933393, 56.1979537])
        assert_equal(qf2.unit, units.m*units.m*units.g)
        assert_equal(i, id(qf2))
        qf2 *= units.K
        assert_almost_equal(qf2.nominal, [350, 525, 700])
        assert_almost_equal(qf2.uncertainty, [20.1677961, 38.0933393, 56.1979537])
        assert_equal(qf2.unit, units.m*units.m*units.g*units.K)
        assert_equal(i, id(qf2))


class Test_QFloat_Div:
    def test_qfloat_math_divmod_single(self):
        qf1 = QFloat(20, 0.5, 's')
        qf2 = QFloat(100, 0.1, 'm')
        qf3 = QFloat(2030, 1, 'cm')
        qf4 = QFloat(5, 0.01, None)

        res1 = qf2 / qf1
        assert_equal(res1.nominal, 5)
        assert_almost_equal(res1.uncertainty, 0.12509996003196805)
        assert_equal(res1.unit, units.m/units.s)
        res1f = qf2 // qf1
        assert_equal(res1f.nominal, 5)
        assert_almost_equal(res1f.uncertainty, 0.0)
        assert_equal(res1f.unit, units.m/units.s)
        res1m = qf2 % qf1
        assert_equal(res1m.nominal, 0)
        # assert_almost_equal(res1m.uncertainty, np.inf)
        assert_equal(res1m.unit, units.m)
        # inverse inverse
        res2 = qf1 / qf2
        assert_equal(res2.nominal, 0.2)
        assert_almost_equal(res2.uncertainty, 0.005003998401278721)
        assert_equal(res2.unit, units.s/units.m)
        res2f = qf1 // qf2
        assert_equal(res2f.nominal, 0)
        assert_almost_equal(res2f.uncertainty, 0.0)
        assert_equal(res2f.unit, units.s/units.m)
        res2m = qf1 % qf2
        assert_equal(res2m.nominal, 20)
        assert_almost_equal(res2m.uncertainty, 0.5)
        assert_equal(res2m.unit, units.s)

        # same dimensionality
        res3 = qf3 / qf2
        assert_equal(res3.nominal, 20.3)
        assert_almost_equal(res3.uncertainty, 0.022629405648403586)
        assert_equal(res3.unit, units.cm/units.m)
        res3f = qf3 // qf2
        assert_equal(res3f.nominal, 20)
        assert_almost_equal(res3f.uncertainty, 0.0)
        assert_equal(res3f.unit, units.cm/units.m)
        res3m = qf3 % qf2
        assert_equal(res3m.nominal, 30)
        assert_almost_equal(res3m.uncertainty, 2.23606797749979)
        assert_equal(res3m.unit, units.cm)

        # with no dimensionality
        res4 = qf4 / qf2
        assert_equal(res4.nominal, 0.05)
        assert_almost_equal(res4.uncertainty, 0.00011180339887498949)
        assert_equal(res4.unit, units.Unit('1/m'))
        res4f = qf4 // qf2
        assert_equal(res4f.nominal, 0)
        assert_almost_equal(res4f.uncertainty, 0.0)
        assert_equal(res4f.unit, units.Unit('1/m'))
        res4m = qf4 % qf2
        assert_equal(res4m.nominal, 5)
        assert_almost_equal(res4m.uncertainty, 0.01)
        assert_equal(res4m.unit, units.dimensionless_unscaled)

        # with numbers
        res5 = qf1 / 7
        assert_almost_equal(res5.nominal, 2.857142857142857)
        assert_almost_equal(res5.uncertainty, 0.07142857142857142)
        assert_equal(res5.unit, units.s)
        res5f = qf1 // 7
        assert_equal(res5f.nominal, 2)
        assert_almost_equal(res5f.uncertainty, 0.0)
        assert_equal(res5f.unit, units.s)
        res5m = qf1 % 7
        assert_equal(res5m.nominal, 6)
        assert_almost_equal(res5m.uncertainty, 0.5)
        assert_equal(res5m.unit, units.s)
        # and inverse
        res6 = 70 / qf1
        assert_almost_equal(res6.nominal, 3.5)
        assert_almost_equal(res6.uncertainty, 0.0875)
        assert_equal(res6.unit, units.Unit('1/s'))
        res6f = 70 // qf1
        assert_almost_equal(res6f.nominal, 3.0)
        assert_almost_equal(res6f.uncertainty, 0.0)
        assert_equal(res6f.unit, units.Unit('1/s'))
        res6m = 70 % qf1
        assert_almost_equal(res6m.nominal, 10)
        assert_almost_equal(res6m.uncertainty, 1.5)
        assert_equal(res6m.unit, units.dimensionless_unscaled)

        # with units
        res7 = qf1 / units.m
        assert_equal(res7.nominal, qf1.nominal)
        assert_equal(res7.uncertainty, qf1.uncertainty)
        assert_equal(res7.unit, units.s/units.m)
        res7m = qf1 % units.m
        assert_equal(res7m.unit, units.s)
        # string
        res8 = qf1 / 'm'
        assert_equal(res8.nominal, qf1.nominal)
        assert_equal(res8.uncertainty, qf1.uncertainty)
        assert_equal(res8.unit, units.s/units.m)
        res8m = qf1 % 'm'
        assert_equal(res8m.unit, units.s)

    def test_qfloat_math_divmod_array(self):
        qf1 = QFloat(np.arange(1, 5)*2, np.arange(1, 5)*0.01, 'm')
        qf2 = QFloat(np.arange(1, 5), np.arange(1, 5)*0.01, 's')
        qf3 = QFloat(2, 0.1, 'min')
        qf4 = QFloat(np.arange(1, 17).reshape((4, 4)),
                     np.arange(1, 17).reshape((4, 4))*0.01, 'km')
        qf5 = QFloat(np.arange(1, 17).reshape((4, 4))*4.5,
                     np.arange(1, 17).reshape((4, 4))*0.01, 'h')

        res1 = qf1 / qf2
        assert_equal(res1.nominal, np.ones(4)*2)
        assert_almost_equal(res1.uncertainty, np.ones(4)*0.022360679774997897)
        assert_equal(res1.unit, units.m/units.s)
        res1f = qf1 // qf2
        assert_equal(res1f.nominal, np.ones(4)*2)
        assert_equal(res1f.uncertainty, np.ones(4)*0.0)
        assert_equal(res1f.unit, units.m/units.s)
        res1m = qf1 % qf2
        assert_equal(res1m.nominal, np.ones(4)*0)
        # not continuous
        # assert_equal(res1m.uncertainty, np.ones(4)*0.0)
        assert_equal(res1m.unit, units.m)
        # inverse
        res2 = qf2 / qf1
        assert_equal(res2.nominal, np.ones(4)*0.5)
        assert_almost_equal(res2.uncertainty, np.ones(4)*0.005590169943749474)
        assert_equal(res2.unit, units.s/units.m)
        res2f = qf2 // qf1
        assert_equal(res2f.nominal, np.ones(4)*0)
        assert_equal(res2f.uncertainty, np.ones(4)*0.0)
        assert_equal(res2f.unit, units.s/units.m)
        res2m = qf2 % qf1
        assert_equal(res2m.nominal, [1, 2, 3, 4])
        assert_equal(res2m.uncertainty, [0.01, 0.02, 0.03, 0.04])
        assert_equal(res2m.unit, units.s)

        # 2D arrays
        res3 = qf5 / qf4
        assert_equal(res3.nominal, np.ones(16).reshape((4, 4))*4.5)
        assert_almost_equal(res3.uncertainty,
                            np.ones(16).reshape((4, 4))*0.046097722286464436)
        assert_equal(res3.unit, units.h/units.km)
        res4 = qf5 // qf4
        assert_equal(res4.nominal, np.ones(16).reshape((4, 4))*4.0)
        assert_almost_equal(res4.uncertainty,
                            np.ones(16).reshape((4, 4))*0.0)
        assert_equal(res4.unit, units.h/units.km)
        res4 = qf5 % qf4
        assert_equal(res4.nominal, np.arange(1, 17).reshape((4, 4))*0.5)
        assert_almost_equal(res4.uncertainty,
                            np.arange(1, 17).reshape((4, 4))*0.04123105625617)
        assert_equal(res4.unit, units.h)

        # Array and single
        res5 = qf2 / qf3
        assert_equal(res5.nominal, np.arange(1, 5)*0.5)
        assert_almost_equal(res5.uncertainty,
                            np.arange(1, 5)*0.025495097567963927)
        assert_equal(res5.unit, units.s/units.min)
        res5f = qf2 // qf3
        assert_equal(res5f.nominal, [0, 1, 1, 2])
        assert_almost_equal(res5f.uncertainty, np.zeros(4))
        assert_equal(res5f.unit, units.s/units.min)
        res5m = qf2 % qf3
        assert_equal(res5m.nominal, [1, 0, 1, 0])
        assert_almost_equal(res5m.uncertainty, [0.01, np.inf, 0.1044031, np.inf])
        assert_equal(res5m.unit, units.s)
        # inverse
        res6 = qf3 / qf2
        assert_almost_equal(res6.nominal, [2, 1, 2/3, 0.5])
        assert_almost_equal(res6.uncertainty, [0.10198039027185571,
                                               0.050990195135927854,
                                               0.033993463423951896,
                                               0.025495097567963927])
        assert_equal(res6.unit, units.min/units.s)
        res6f = qf3 // qf2
        assert_equal(res6f.nominal, [2, 1, 0, 0])
        assert_almost_equal(res6f.uncertainty, np.zeros(4))
        assert_equal(res6f.unit, units.min/units.s)
        res6m = qf3 % qf2
        assert_equal(res6m.nominal, [0, 0, 2, 2])
        assert_equal(res6m.unit, units.min)

        # with units
        res7 = qf2 / units.m
        assert_almost_equal(res7.nominal, qf2.nominal)
        assert_almost_equal(res7.uncertainty, qf2.uncertainty)
        assert_equal(res7.unit, units.s/units.m)
        res7m = qf2 % units.m
        assert_equal(res7m.unit, units.s)
        # string
        res8 = qf2 / 'm'
        assert_almost_equal(res8.nominal, qf2.nominal)
        assert_almost_equal(res8.uncertainty, qf2.uncertainty)
        assert_equal(res8.unit, units.s/units.m)
        res8m = qf2 % 'm'
        assert_equal(res8m.unit, units.s)

        # with numbers
        res9 = qf1 / 4
        assert_almost_equal(res9.nominal, np.arange(1, 5)*0.5)
        assert_almost_equal(res9.uncertainty, np.arange(1, 5)*0.0025)
        assert_equal(res9.unit, units.m)
        res9f = qf1 // 4
        assert_almost_equal(res9f.nominal, [0, 1, 1, 2])
        assert_almost_equal(res9f.uncertainty, [0, 0, 0, 0])
        assert_equal(res9f.unit, units.m)
        res9m = qf1 % 4
        assert_almost_equal(res9m.nominal, [2, 0, 2, 0])
        assert_almost_equal(res9m.uncertainty, [0.01, np.nan, 0.03, np.nan])
        assert_equal(res9m.unit, units.m)
        res_a = 8 / qf1
        assert_almost_equal(res_a.nominal, [4, 2, 4/3, 1])
        assert_almost_equal(res_a.uncertainty, [0.02, 0.01, 6/900,
                                                0.005])
        assert_equal(res_a.unit, units.Unit('1/m'))
        res_af = 8 // qf1
        assert_almost_equal(res_af.nominal, [4, 2, 1, 1])
        assert_almost_equal(res_af.uncertainty, [0.0, 0.0, 0.0, 0.0])
        assert_equal(res_af.unit, units.Unit('1/m'))
        res_am = 8 % qf1
        assert_almost_equal(res_am.nominal, [0, 0, 2, 0])
        assert_almost_equal(res_am.std_dev, [np.inf, np.inf, 0.03, np.inf])
        assert_equal(res_am.unit, units.dimensionless_unscaled)

    def test_qfloat_math_truediv_inline(self):
        qf1 = QFloat(30, 0.5, 's')
        qf2 = QFloat(12000, 20, 'm')
        qf3 = QFloat([6000, 3000, 150], [15, 10, 0.5], 'kg')
        qf4 = QFloat([10, 20, 0.1], [0.1, 0.4, 0.01])

        # qf/qf
        i = id(qf2)
        qf2 /= qf1
        assert_equal(qf2.nominal, 400)
        assert_almost_equal(qf2.uncertainty, 6.699917080747261)
        assert_equal(qf2.unit, units.m/units.s)
        assert_equal(i, id(qf2))
        # number
        qf2 /= 2
        assert_equal(qf2.nominal, 200)
        assert_almost_equal(qf2.uncertainty, 3.3499585403736303)
        assert_equal(qf2.unit, units.m/units.s)
        assert_equal(i, id(qf2))
        # unit
        qf2 /= units.s
        assert_equal(qf2.nominal, 200)
        assert_almost_equal(qf2.uncertainty, 3.3499585403736303)
        assert_equal(qf2.unit, units.m/(units.s*units.s))
        assert_equal(i, id(qf2))
        # string
        qf2 /= 'kg'
        assert_equal(qf2.nominal, 200)
        assert_almost_equal(qf2.uncertainty, 3.3499585403736303)
        assert_equal(qf2.unit, units.m/(units.s*units.s*units.kg))
        assert_equal(i, id(qf2))
        # quantity
        qf2 /= 4/units.s
        assert_equal(qf2.nominal, 50)
        assert_almost_equal(qf2.uncertainty, 0.8374896350934076)
        assert_equal(qf2.unit, units.m/(units.s*units.kg))
        assert_equal(i, id(qf2))

        # array / qfloat
        i = id(qf3)
        qf3 /= qf1
        assert_almost_equal(qf3.nominal, [200, 100, 5])
        assert_almost_equal(qf3.uncertainty, [3.370624736026114,
                                              1.699673171197595,
                                              0.08498365855987974])
        assert_equal(qf3.unit, units.kg/units.s)
        assert_equal(i, id(qf3))
        # array / array
        qf3 /= qf4
        assert_almost_equal(qf3.nominal, [20, 5, 50])
        assert_almost_equal(qf3.uncertainty, [0.39193253387682825,
                                              0.13123346456686352,
                                              5.071708018234312])
        assert_equal(qf3.unit, units.kg/units.s)
        assert_equal(i, id(qf3))
        # array / number
        qf3 /= 5
        assert_almost_equal(qf3.nominal, [4, 1, 10])
        assert_almost_equal(qf3.uncertainty, [0.07838650677536566,
                                              0.026246692913372706,
                                              1.0143416036468624])
        assert_equal(qf3.unit, units.kg/units.s)
        assert_equal(i, id(qf3))
        # array  / unit
        qf3 /= units.m
        assert_almost_equal(qf3.nominal, [4, 1, 10])
        assert_almost_equal(qf3.uncertainty, [0.07838650677536566,
                                              0.026246692913372706,
                                              1.0143416036468624])
        assert_equal(qf3.unit, units.kg/(units.s*units.m))
        assert_equal(i, id(qf3))
        # array / string
        qf3 /= 's'
        assert_almost_equal(qf3.nominal, [4, 1, 10])
        assert_almost_equal(qf3.uncertainty, [0.07838650677536566,
                                              0.026246692913372706,
                                              1.0143416036468624])
        assert_equal(qf3.unit, units.kg/(units.s*units.s*units.m))
        assert_equal(i, id(qf3))
        # array / quantity
        qf3 /= 1/units.kg
        assert_almost_equal(qf3.nominal, [4, 1, 10])
        assert_almost_equal(qf3.uncertainty, [0.07838650677536566,
                                              0.026246692913372706,
                                              1.0143416036468624])
        assert_equal(qf3.unit, units.kg*units.kg/(units.s*units.s*units.m))
        assert_equal(i, id(qf3))

    def test_qfloat_math_floordiv_inline(self):
        qf1 = QFloat(30, 0.5, 's')
        qf2 = QFloat(12010, 20, 'm')
        qf3 = QFloat([6040, 3012, 153], [15, 10, 0.5], 'kg')
        qf4 = QFloat([10, 20, 0.1], [0.1, 0.4, 0.01])

        # qf/qf
        i = id(qf2)
        qf2 //= qf1
        assert_equal(qf2.nominal, 400)
        assert_almost_equal(qf2.uncertainty, 0)
        assert_equal(qf2.unit, units.m/units.s)
        assert_equal(i, id(qf2))
        # number
        qf2 //= 7
        assert_equal(qf2.nominal, 57)
        assert_almost_equal(qf2.uncertainty, 0)
        assert_equal(qf2.unit, units.m/units.s)
        assert_equal(i, id(qf2))
        # unit
        qf2 //= units.s
        assert_equal(qf2.nominal, 57)
        assert_almost_equal(qf2.uncertainty, 0)
        assert_equal(qf2.unit, units.m/(units.s*units.s))
        assert_equal(i, id(qf2))
        # string
        qf2 //= 'kg'
        assert_equal(qf2.nominal, 57)
        assert_almost_equal(qf2.uncertainty, 0)
        assert_equal(qf2.unit, units.m/(units.s*units.s*units.kg))
        assert_equal(i, id(qf2))
        # quantity
        qf2 //= 4/units.s
        assert_equal(qf2.nominal, 14)
        assert_almost_equal(qf2.uncertainty, 0.0)
        assert_equal(qf2.unit, units.m/(units.s*units.kg))
        assert_equal(i, id(qf2))

        # array / qfloat
        i = id(qf3)
        qf3 //= qf1
        assert_almost_equal(qf3.nominal, [201, 100, 5])
        assert_almost_equal(qf3.uncertainty, [0, 0, 0])
        assert_equal(qf3.unit, units.kg/units.s)
        assert_equal(i, id(qf3))
        # array / array
        qf3 //= qf4
        assert_almost_equal(qf3.nominal, [20, 5, 49])
        assert_almost_equal(qf3.uncertainty, [0, 0, 0])
        assert_equal(qf3.unit, units.kg/units.s)
        assert_equal(i, id(qf3))
        # array / number
        qf3 //= 5
        assert_almost_equal(qf3.nominal, [4, 1, 9])
        assert_almost_equal(qf3.uncertainty, [0, 0, 0])
        assert_equal(qf3.unit, units.kg/units.s)
        assert_equal(i, id(qf3))
        # array  / unit
        qf3 //= units.m
        assert_almost_equal(qf3.nominal, [4, 1, 9])
        assert_almost_equal(qf3.uncertainty, [0, 0, 0])
        assert_equal(qf3.unit, units.kg/(units.s*units.m))
        assert_equal(i, id(qf3))
        # array / string
        qf3 //= 's'
        assert_almost_equal(qf3.nominal, [4, 1, 9])
        assert_almost_equal(qf3.uncertainty, [0, 0, 0])
        assert_equal(qf3.unit, units.kg/(units.s*units.s*units.m))
        assert_equal(i, id(qf3))
        # array / quantity
        qf3 //= 1/units.kg
        assert_almost_equal(qf3.nominal, [4, 1, 9])
        assert_almost_equal(qf3.uncertainty, [0, 0, 0])
        assert_equal(qf3.unit, units.kg*units.kg/(units.s*units.s*units.m))
        assert_equal(i, id(qf3))

    def test_qfloat_math_mod_inline(self):
        qf1 = QFloat(30, 0.5, 's')
        qf2 = QFloat(12010, 20, 'm')
        qf3 = QFloat([6040, 3012, 153], [15, 10, 0.5], 'kg')
        qf4 = QFloat([7, 5, 2], [0.1, 0.4, 0.01])

        # crazy uncertainties

        # qf/qf
        i = id(qf2)
        qf2 %= qf1
        assert_equal(qf2.nominal, 10)
        assert_equal(qf2.unit, units.m)
        assert_equal(i, id(qf2))
        # number
        qf2 %= 7
        assert_equal(qf2.nominal, 3)
        assert_equal(qf2.unit, units.m)
        assert_equal(i, id(qf2))
        # quantity
        qf2 %= 2/units.s
        assert_equal(qf2.nominal, 1)
        assert_equal(qf2.unit, units.m)
        assert_equal(i, id(qf2))

        # array / qfloat
        i = id(qf3)
        qf3 %= qf1
        assert_almost_equal(qf3.nominal, [10, 12, 3])
        assert_equal(qf3.unit, units.kg)
        assert_equal(i, id(qf3))
        # array / array
        qf3 %= qf4
        assert_almost_equal(qf3.nominal, [3, 2, 1])
        assert_equal(qf3.unit, units.kg)
        assert_equal(i, id(qf3))
        # array / number
        qf3 %= 2
        assert_almost_equal(qf3.nominal, [1, 0, 1])
        assert_equal(qf3.unit, units.kg)
        assert_equal(i, id(qf3))
        # array / quantity
        qf3 %= 1/units.kg
        assert_almost_equal(qf3.nominal, [0, 0, 0])
        assert_equal(qf3.unit, units.kg)
        assert_equal(i, id(qf3))


class Test_QFloat_Pow:
    def test_qfloat_math_power_single(self):
        qf1 = QFloat(2.0, 0.1, 'm')
        qf2 = QFloat(1.5, 0.2, 'cm')
        qf3 = QFloat(0.5, 0.01)
        qf4 = QFloat(4.0, 0.01)

        with pytest.raises(ValueError):
            qf1 ** qf2

        res1 = qf1 ** qf3
        assert_almost_equal(res1.nominal, np.sqrt(2))
        assert_almost_equal(res1.uncertainty, 0.03668910741328604)
        assert_equal(res1.unit, units.m**0.5)

        res2 = qf1 ** qf4
        assert_almost_equal(res2.nominal, 2**4)
        assert_almost_equal(res2.uncertainty, 3.201921235314246)
        assert_equal(res2.unit, units.m*units.m*units.m*units.m)

        res3 = qf1 ** 2
        assert_almost_equal(res3.nominal, 4)
        assert_almost_equal(res3.uncertainty, 0.4)
        assert_equal(res3.unit, units.m*units.m)

        res4 = qf1 ** 0.5
        assert_almost_equal(res4.nominal, np.sqrt(2))
        assert_almost_equal(res4.uncertainty, 0.03535533905932738)
        assert_equal(res4.unit, units.m**0.5)

        # only dimensionless quantity
        with pytest.raises(ValueError):
            qf1 ** (1*units.m)
        res5 = qf1 ** (1*units.dimensionless_unscaled)
        assert_almost_equal(res5.nominal, 2)
        assert_almost_equal(res5.uncertainty, 0.1)
        assert_equal(res5.unit, units.m)

        # string and unit must fail, except for dimensionless
        with pytest.raises(ValueError):
            qf1 ** 's'
        with pytest.raises(ValueError):
            qf1 ** units.s
        res6 = qf1 ** ''
        assert_almost_equal(res6.nominal, 2)
        assert_almost_equal(res6.uncertainty, 0.1)
        assert_equal(res6.unit, units.m)
        res7 = qf1 ** units.dimensionless_unscaled
        assert_almost_equal(res7.nominal, 2)
        assert_almost_equal(res7.uncertainty, 0.1)
        assert_equal(res7.unit, units.m)

        # only dimensionless should power
        with pytest.raises(ValueError):
            1 ** qf1
        res8 = 2 ** qf4
        assert_almost_equal(res8.nominal, 16)
        assert_almost_equal(res8.uncertainty, 0.11090354888959125)
        assert_equal(res8.unit, units.dimensionless_unscaled)

    def test_qfloat_math_power_array(self):
        qf1 = QFloat([2, 3, 4], [0.1, 0.2, 0.3], 'm')
        qf2 = QFloat(1.5, 0.2, 'cm')
        qf3 = QFloat(0.5, 0.01)
        qf4 = QFloat([1, 2, 3], [0.1, 0.2, 0.3])

        with pytest.raises(ValueError):
            qf1 ** qf2

        res1 = qf1 ** qf3
        assert_almost_equal(res1.nominal, np.sqrt([2, 3, 4]))
        assert_almost_equal(res1.uncertainty, [0.03668910741328604,
                                               0.06078995000472617,
                                               0.07996077052073174])
        assert_equal(res1.unit, units.m**0.5)

        with pytest.raises(ValueError):
            qf1 ** qf4

        res3 = qf1 ** 2
        assert_almost_equal(res3.nominal, [4, 9, 16])
        assert_almost_equal(res3.uncertainty, [0.4, 1.2, 2.4])
        assert_equal(res3.unit, units.m*units.m)

        res4 = qf1 ** 0.5
        assert_almost_equal(res4.nominal, np.sqrt([2, 3, 4]))
        assert_almost_equal(res4.uncertainty, [0.03535533905932738,
                                               0.057735026918962574,
                                               0.075])
        assert_equal(res4.unit, units.m**0.5)

        # only dimensionless quantity
        with pytest.raises(ValueError):
            qf1 ** (1*units.m)
        res5 = qf1 ** (1*units.dimensionless_unscaled)
        assert_almost_equal(res5.nominal, [2, 3, 4])
        assert_almost_equal(res5.uncertainty, [0.1, 0.2, 0.3])
        assert_equal(res5.unit, units.m)

        # string and unit must fail, except for dimensionless
        with pytest.raises(ValueError):
            qf1 ** 's'
        with pytest.raises(ValueError):
            qf1 ** units.s
        res6 = qf1 ** ''
        assert_almost_equal(res5.nominal, [2, 3, 4])
        assert_almost_equal(res5.uncertainty, [0.1, 0.2, 0.3])
        assert_equal(res6.unit, units.m)
        res7 = qf1 ** units.dimensionless_unscaled
        assert_almost_equal(res5.nominal, [2, 3, 4])
        assert_almost_equal(res5.uncertainty, [0.1, 0.2, 0.3])
        assert_equal(res7.unit, units.m)

        # only dimensionless size-1 array should work
        with pytest.raises(ValueError):
            1 ** qf1
        with pytest.raises(ValueError):
            1 ** qf2
        with pytest.raises(ValueError):
            1 ** qf4

    def test_qfloat_math_power_inline(self):
        qf1 = QFloat(2.0, 0.1, 'm')
        qf2 = QFloat(1.5, 0.2, 'cm')
        qf3 = QFloat(0.5, 0.01)
        qf4 = QFloat(4.0, 0.01)
        qf5 = QFloat([2, 3, 4], [0.1, 0.2, 0.3], 'm')
        qf6 = QFloat([1, 2, 3], [0.1, 0.2, 0.3])

        i = id(qf1)
        with pytest.raises(ValueError):
            qf1 **= qf2
        with pytest.raises(ValueError):
            qf1 **= qf5
        with pytest.raises(ValueError):
            qf1 **= qf6
        with pytest.raises(ValueError):
            qf1 **= 'm'
        with pytest.raises(ValueError):
            qf1 **= units.s
        assert_equal(i, id(qf1))
        qf1 **= qf3
        assert_almost_equal(qf1.nominal, np.sqrt(2))
        assert_almost_equal(qf1.uncertainty, 0.03668910741328604)
        assert_equal(qf1.unit, units.m**0.5)
        assert_equal(i, id(qf1))
        qf1 **= qf4
        assert_almost_equal(qf1.nominal, 4)
        assert_almost_equal(qf1.uncertainty, 0.4153212953387695)
        assert_equal(qf1.unit, units.m**2)
        assert_equal(i, id(qf1))
        qf1 **= 3
        assert_almost_equal(qf1.nominal, 64)
        assert_almost_equal(qf1.uncertainty, 19.935422176260943)
        assert_equal(qf1.unit, units.m**6)
        assert_equal(i, id(qf1))
        qf1 **= ''
        assert_almost_equal(qf1.nominal, 64)
        assert_almost_equal(qf1.uncertainty, 19.935422176260943)
        assert_equal(qf1.unit, units.m**6)
        assert_equal(i, id(qf1))
        qf1 **= units.dimensionless_unscaled
        assert_almost_equal(qf1.nominal, 64)
        assert_almost_equal(qf1.uncertainty, 19.935422176260943)
        assert_equal(qf1.unit, units.m**6)
        assert_equal(i, id(qf1))

        i = id(qf5)
        with pytest.raises(ValueError):
            qf5 **= qf1
        with pytest.raises(ValueError):
            qf5 **= qf2
        with pytest.raises(ValueError):
            qf5 **= qf6
        with pytest.raises(ValueError):
            qf5 **= 'm'
        with pytest.raises(ValueError):
            qf5 **= units.s
        assert_equal(i, id(qf5))
        qf5 **= qf3
        assert_almost_equal(qf5.nominal, np.sqrt([2, 3, 4]))
        assert_almost_equal(qf5.uncertainty, [0.03668910741328604,
                                              0.06078995000472617,
                                              0.07996077052073174])
        assert_equal(qf5.unit, units.m**0.5)
        assert_equal(i, id(qf5))
        qf5 **= qf4
        assert_almost_equal(qf5.nominal, np.square([2, 3, 4]))
        assert_almost_equal(qf5.uncertainty, [0.4153212953387695,
                                              1.2644622006872943,
                                              2.5611469725808176])
        assert_equal(qf5.unit, units.m**2)
        assert_equal(i, id(qf5))
        qf5 **= 3
        assert_almost_equal(qf5.nominal, [64, 729, 4096])
        assert_almost_equal(qf5.uncertainty, [19.935422176260943,
                                              307.2643147670124,
                                              1966.960874942068])
        assert_equal(qf5.unit, units.m**6)
        assert_equal(i, id(qf5))
        qf5 **= ''
        assert_almost_equal(qf5.nominal, [64, 729, 4096])
        assert_almost_equal(qf5.uncertainty, [19.935422176260943,
                                              307.2643147670124,
                                              1966.960874942068])
        assert_equal(qf5.unit, units.m**6)
        assert_equal(i, id(qf5))
        qf5 **= units.dimensionless_unscaled
        assert_almost_equal(qf5.nominal, [64, 729, 4096])
        assert_almost_equal(qf5.uncertainty, [19.935422176260943,
                                              307.2643147670124,
                                              1966.960874942068])
        assert_equal(qf5.unit, units.m**6)
        assert_equal(i, id(qf5))


class Test_QFloat_PosNeg:
    def test_qfloat_math_neg(self):
        qf1 = QFloat(1.0, 0.1, 'm')
        qfn1 = -qf1
        assert_equal(qfn1.nominal, -1.0)
        assert_equal(qfn1.uncertainty, 0.1)
        assert_equal(qfn1.unit, units.m)

        qf2 = QFloat(-5, 0.1, 'm')
        qfn2 = -qf2
        assert_equal(qfn2.nominal, 5.0)
        assert_equal(qfn2.uncertainty, 0.1)
        assert_equal(qfn2.unit, units.m)

        qf3 = QFloat([-2, 3, -5, 10], [0.1, 0.4, 0.3, -0.1], 'm')
        qfn3 = -qf3
        assert_equal(qfn3.nominal, [2, -3, 5, -10])
        assert_equal(qfn3.uncertainty, [0.1, 0.4, 0.3, 0.1])
        assert_equal(qfn3.unit, units.m)

    def test_qfloat_math_pos(self):
        qf1 = QFloat(1.0, 0.1, 'm')
        qfn1 = +qf1
        assert_equal(qfn1.nominal, 1.0)
        assert_equal(qfn1.uncertainty, 0.1)
        assert_equal(qfn1.unit, units.m)

        qf2 = QFloat(-5, 0.1, 'm')
        qfn2 = +qf2
        assert_equal(qfn2.nominal, -5.0)
        assert_equal(qfn2.uncertainty, 0.1)
        assert_equal(qfn2.unit, units.m)

        qf3 = QFloat([-2, 3, -5, 10], [0.1, 0.4, 0.3, -0.1], 'm')
        qfn3 = +qf3
        assert_equal(qfn3.nominal, [-2, 3, -5, 10])
        assert_equal(qfn3.uncertainty, [0.1, 0.4, 0.3, 0.1])
        assert_equal(qfn3.unit, units.m)

    def test_qfloat_math_abs(self):
        qf1 = QFloat(1.0, 0.1, 'm')
        qfn1 = abs(qf1)
        assert_equal(qfn1.nominal, 1.0)
        assert_equal(qfn1.uncertainty, 0.1)
        assert_equal(qfn1.unit, units.m)

        qf2 = QFloat(-5, 0.1, 'm')
        qfn2 = abs(qf2)
        assert_equal(qfn2.nominal, 5.0)
        assert_equal(qfn2.uncertainty, 0.1)
        assert_equal(qfn2.unit, units.m)

        qf3 = QFloat([-2, 3, -5, 10], [0.1, 0.4, 0.3, -0.1], 'm')
        qfn3 = abs(qf3)
        assert_equal(qfn3.nominal, [2, 3, 5, 10])
        assert_equal(qfn3.uncertainty, [0.1, 0.4, 0.3, 0.1])
        assert_equal(qfn3.unit, units.m)


class Test_QFloat_Types:
    def test_qfloat_math_float(self):
        qf1 = QFloat(1.5, 0.4, 'm')
        qf2 = QFloat([1.4, 2.5, 3.6], [0.2, 0.3, 0.4], 's')

        assert_equal(float(qf1), 1.5)
        with pytest.raises(TypeError):
            float(qf2)

    def test_qfloat_math_int(self):
        qf1 = QFloat(1.5, 0.4, 'm')
        qf2 = QFloat([1.4, 2.5, 3.6], [0.2, 0.3, 0.4], 's')

        assert_equal(int(qf1), 1)
        with pytest.raises(TypeError):
            int(qf2)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Custom frame class to support memmapping."""

# Important: We reimplemented and renamed astropy's frame to better
# handle disk memmap and an easier unit/uncertainty workflow

import os
import numpy as np
import copy as cp
import tempfile
from astropy import units as u
from astropy.io import fits
from astropy.wcs import WCS

from ..py_utils import check_iterable
from .memmap import MemMapArray
from .compat import _to_ccddata, _to_hdu, _merge_and_clean_header, _write_fits
from .._unit_property import unit_property


__all__ = ['FrameData']


def _get_shape(d):
    if hasattr(d, 'shape'):
        ds = d.shape
    else:
        ds = np.array(d).shape
    return ds


def shape_consistency(data=None, uncertainty=None, mask=None):
    """Check shape consistency across `data`, `uncertaitny` and `mask`."""
    if data is None and uncertainty is not None:
        raise ValueError('Uncertainty set for an empty data.')
    if data is None and mask not in (None, False):
        raise ValueError('Mask set for an empty data.')

    dshape = _get_shape(data)

    if uncertainty is not None:
        ushape = _get_shape(uncertainty)

        if ushape == ():
            uncertainty = np.ones(dshape)*uncertainty
            ushape = uncertainty.shape

        if ushape != dshape:
            raise ValueError(f'Uncertainty shape {ushape} don\'t match'
                             f' Data shape {dshape}.')

    if mask is not None:
        mshape = _get_shape(mask)

        if mshape == ():
            mask = np.logical_or(np.zeros(dshape), mask)
            mshape = mask.shape

        if mask.shape != dshape:
            raise ValueError(f'Mask shape {mshape} don\'t match'
                             f' Data shape {dshape}.')

    return data, uncertainty, mask


def extract_units(data, unit):
    """Extract and compare units if they are consistent."""
    if hasattr(data, 'unit'):
        dunit = u.Unit(data.unit)
    else:
        dunit = None
    if unit is not None:
        unit = u.Unit(unit)
    else:
        unit = None

    if dunit is not None and unit is not None:
        if dunit is not unit:
            raise ValueError(f"Unit {unit} cannot be set for a data with"
                             f" unit {dunit}")

    # if dunit is None, assign unit to it.
    if dunit is None:
        dunit = unit

    return dunit


def uncertainty_unit_consistency(unit, uncertainty):
    """Check if an uncertainty can be set based on its unit.

    Parameters
    ----------
    unit: string or `~astropy.units.Unit`
        Target uncertainty unit. Mainly, this is the FrameData unit.
    uncertainty: array_like
        Uncertainty object. If it has units, it will be converted to the
        target.

    Returns
    -------
    uncertainty: `numpy.ndarray`
        Uncertainty converted to the target unit.
    """
    # Assume a plain array uncertainty if no unit.
    if not hasattr(uncertainty, 'unit'):
        return np.array(uncertainty)

    u_unit = uncertainty.unit
    if u_unit == unit:
        return np.array(uncertainty)

    return u.Quantity(uncertainty, u_unit).to(unit).value


def setup_filename(frame, cache_folder=None, filename=None):
    """Handle filename and cache folder to a frame."""
    if not isinstance(frame, FrameData):
        raise ValueError('Only FrameData accepted.')

    cache_folder_ccd = frame.cache_folder
    filename_ccd = frame.cache_filename

    # explicit set must be over defult
    filename = filename or filename_ccd
    if filename is None:
        filename = tempfile.NamedTemporaryFile(suffix='.npy').name

    if cache_folder is None and \
       os.path.dirname(filename) not in ('', tempfile.gettempdir()):
        # we need filename dir for cache_folder
        cache_folder = os.path.dirname(filename)

    filename = os.path.basename(filename)

    # explicit set must be over defult
    cache_folder = cache_folder or cache_folder_ccd
    if cache_folder is None:
        cache_folder = tempfile.TemporaryDirectory(prefix='astropop_').name

    os.makedirs(cache_folder, exist_ok=True)

    frame.cache_folder = cache_folder
    frame.cache_filename = filename

    return os.path.join(cache_folder, filename)


@unit_property
class FrameData:
    """Data container for image frame to handle memmapping data from disk.

    The main difference from Astropy's `~astropy.nddata.CCDData` is the
    memmapping itself. However it handles uncertainties in a totally different
    way. It stores only StdDev uncertainty arrays. It also stores the unit.

    Parameters
    ----------
    data : array_like or `~astropy.units.Quantity`
        The main data values. If `~astropy.units.Quantity`, unit will be set
        automatically.
    unit : `~astropy.units.Unit` or string (optional)
        The data unit. Must be `~astropy.units.Unit` compliant.
    dtype : string or `~numpy.dtype` (optional)
        Mandatory dtype of the data.
    uncertainty : array_like or `~astropy.nddata.Uncertanty` or `None` \
                    (optional)
        Uncertainty of the data.
    u_dtype : string or `~numpy.dtype` (optional)
        Mandatory dtype of uncertainty.
    mask : array_like or `None` (optional)
        Frame mask.
    m_dtype : string or `~numpy.dtype` (optional)
        Mandatory dtype of mask. Default `bool`compilant
    wcs : `dict`, `~astropy.fits.Header` or `~astropy.wcs.WCS` (optional)
        World Coordinate System of the image.
    meta or header: `dict` or `astropy.fits.Header` (optional)
        Metadata (header) of the frame. If both set, they will be merged.
        `header` priority.
    cache_folder : string, `~pathlib.Path` or `None` (optional)
        Place to store the cached `FrameData`
    cache_filename : string, `~pathlib.Path` or `None` (optional)
        Base file name to store the cached `FrameData`.
    use_memmap_backend : `bool` (optional)
        True if enable memmap in constructor.

    Notes
    -----
    - The physical unit is assumed to be the same for data and uncertainty.
      So, we droped the support for data with data with different uncertainty
      unit, like `~astropy.nddata.ccddata.CCDData` does.
    - As this is intended to be a safe container for data, it do not handle
      builtin math operations. For math operations using FrameData, check
      `~astropop.ccd_processing.imarith` module.
    """

    # TODO: write_fits

    _memmapping = False
    _unit = None
    _data = None
    _mask = None
    _unct = None
    _wcs = None
    _meta = None
    _origin = None
    _history = None
    _comments = None

    def __init__(self, data, unit=None, dtype=None,
                 uncertainty=None, u_dtype=None,
                 mask=None, m_dtype=bool,
                 wcs=None, meta=None, header=None,
                 cache_folder=None, cache_filename=None,
                 use_memmap_backend=False, origin_filename=None):

        self.cache_folder = cache_folder
        self.cache_filename = cache_filename
        self._origin = origin_filename

        # Setup MemMapArray instances
        cache_file = setup_filename(self, self.cache_folder,
                                    self.cache_filename)
        self._update_cache_files(cache_file)

        # Check for memmapping.
        self._memmapping = False
        if use_memmap_backend:
            self.enable_memmap()

        # Masking handle
        if hasattr(data, 'mask'):
            dmask = data.mask
            if mask is not None:
                mask = np.logical_or(dmask, mask)
            else:
                mask = dmask
        if mask is None:  # Default do not mask anything
            mask = False

        # raise errors if incompatible shapes
        data, uncertainty, mask = shape_consistency(data, uncertainty, mask)
        # raise errors if incompatible units
        unit = extract_units(data, unit)
        self.unit = unit
        self._data.reset_data(data, dtype)
        self._unct.reset_data(uncertainty, u_dtype)
        # Masks can also be flags (uint8)
        self._mask.reset_data(mask, m_dtype)

        # avoiding security problems
        self._history = []
        self._comments = []
        self._meta = fits.Header()
        self._header_update(header, meta, wcs)

    def _header_update(self, header, meta=None, wcs=None):
        # merge header and meta. meta with higher priority
        meta, wcs, history, comment = _merge_and_clean_header(meta, header,
                                                              wcs)
        if len(history) > 0:
            self.history = history
        if len(comment) > 0:
            self.comments = comment
        if wcs is not None:
            self._wcs = wcs
        self._meta = meta

    def _update_cache_files(self, cache_file):
        # TODO: if cache folder is changing, remove it
        if self._data is not None:
            self._data.disable_memmap(remove=True)
            nd = self._data._contained
        else:
            nd = None
        if self._unct is not None:
            self._unct.disable_memmap(remove=True)
            nu = self._unct._contained
        else:
            nu = None
        if self._mask is not None:
            self._mask.disable_memmap(remove=True)
            nm = self._mask._contained
        else:
            nm = None

        self._data = MemMapArray(nd, filename=cache_file + '.data')
        self._unct = MemMapArray(nu, filename=cache_file + '.unct')
        self._mask = MemMapArray(nm, filename=cache_file + '.mask')

    @property
    def history(self):
        """Get the FrameData stored history."""
        return self._history

    @history.setter
    def history(self, value):
        if check_iterable(value):
            self._history = self._history + list(value)
        else:
            self._history.append(value)

    @property
    def comment(self):
        """Get the FrameData stored comments."""
        return self._comments

    @comment.setter
    def comment(self, value):
        if check_iterable(value):
            self._comments = self._comments + list(value)
        else:
            self._comments.append(value)

    @property
    def origin_filename(self):
        """Get the original filename of the data."""
        return self._origin

    @property
    def shape(self):
        """Get the data shape following numpy. `FrameData.data.shape`."""
        return self._data.shape

    @property
    def dtype(self):
        """Get the dta type of the data. `FrameData.data.dtype`."""
        return self._data.dtype

    @property
    def size(self):
        """Get the size of the data. `FrameData.data.size`."""
        return self._data.size

    @property
    def wcs(self):
        """Get the World Coordinate System."""
        return self._wcs

    @wcs.setter
    def wcs(self, value):
        if isinstance(value, WCS) or value is None:
            self._wcs = value
        else:
            raise TypeError('wcs setter value must be a WCS instance.')

    @property
    def meta(self):
        """Get the metadata (header) of the frame."""
        return self._meta

    @meta.setter
    def meta(self, value):
        self._header_update(value)

    @property
    def header(self):
        """Get the header (metadata) of the frame."""
        return self._meta

    @property
    def data(self):
        """Get the main data container."""
        return self._data

    @data.setter
    def data(self, value):
        if hasattr(value, 'unit'):
            dunit = extract_units(value, None)
            self.unit = dunit
        self._data.reset_data(value)

    @property
    def uncertainty(self):
        """Get the uncertainty frame container."""
        return self._unct

    @uncertainty.setter
    def uncertainty(self, value):
        if value is None:
            self._unct.reset_data(value)
        else:
            _, value, _ = shape_consistency(self.data, value, None)
            value = uncertainty_unit_consistency(self.unit, value)
            self._unct.reset_data(value)

    def get_uncertainty(self, return_none=True):
        """Get the uncertainty frame in a safer way.

        In some cases, like interfacing with QFloat, the uncertainty cannot be
        None or an empty container. So, in cases like this, is prefered to get
        a whole matrix containing zeroes. This method is responsible for this
        special returns. For non-empty containers, a copy in `~numpy.ndarray`
        container will be returned.

        Parameters
        ----------
        - return_none: bool (optional)
          If True, an empty uncertainty frame will return only None. Else, a
          matrix filled with zeroes will be returned.
          Default: True
        """
        if self._unct.empty and return_none:
            return None
        if self._unct.empty:
            return np.zeros_like(self._data, dtype=self.dtype)
        return np.array(self._unct, dtype=self.dtype)

    @property
    def mask(self):
        """Access mask data."""
        return self._mask

    @mask.setter
    def mask(self, value):
        _, _, value = shape_consistency(self.data, None, value)
        self._mask.reset_data(value)

    def astype(self, dtype):
        """Return a copy of the current FrameData with new dtype in data."""
        return self.copy(dtype)

    def copy(self, dtype=None):
        """Copy the current FrameData to a new instance.

        Parameters
        ----------
        - dtype: `~numpy.dtype` (optional)
            Data type for the copied FrameData. If `None`, the data type will
            be the same as the original Framedata.
            Default: `None`
        """
        data = np.array(self._data) if not self._data.empty else None
        mask = np.array(self._mask) if not self._mask.empty else None
        unct = np.array(self._unct) if not self._unct.empty else None
        unit = self._unit
        wcs = cp.copy(self._wcs)
        meta = fits.Header(self._meta, copy=True)
        hist = cp.copy(self._history)
        comm = cp.copy(self._comments)
        fname = self._origin
        cache_folder = self.cache_folder
        cache_fname = self.cache_filename

        if dtype is not None:
            data = data.astype(dtype) if data is not None else data
            unct = unct.astype(dtype) if unct is not None else unct

        if cache_fname is not None:
            cache_fname = cache_fname + '_copy'

        nframe = FrameData(data, unit=unit, mask=mask, uncertainty=unct,
                           meta=meta, cache_folder=cache_folder,
                           cache_filename=cache_fname, origin_filename=fname)
        nframe.history = hist
        nframe.comments = comm
        nframe.wcs = wcs
        return nframe

    def enable_memmap(self, filename=None, cache_folder=None):
        """Enable array file memmapping.

        Parameters
        ----------
        filename : str, optional
            Custom memmap file name.
        cache_folder : str, optional
            Custom folder to cache data.
        """
        if filename is None and cache_folder is None:
            self._data.enable_memmap()
            self._unct.enable_memmap()
            self._mask.enable_memmap()
        else:
            cache_file = setup_filename(self, cache_folder, filename)
            self._update_cache_files(cache_file)
            self._data.enable_memmap(cache_file + '.data')
            self._mask.enable_memmap(cache_file + '.mask')
            self._unct.enable_memmap(cache_file + '.unct')
        self._memmapping = True

    def disable_memmap(self):
        """Disable frame file memmapping (load to memory)."""
        self._data.disable_memmap(remove=True)
        self._mask.disable_memmap(remove=True)
        self._unct.disable_memmap(remove=True)
        self._memmapping = False

    def to_hdu(self, **kwargs):
        """Generate an HDUList from this FrameData.

        Parameters
        ----------
        hdu_uncertainty: string, optional
            Extension name to store the uncertainty. If None,
            no uncertainty will be stored.
        hdu_mask: string, optional
            Extension name to store the mask. If None, no mask
            will be saved.
        unit_key: string, optional
            Header key for physical unit.
        wcs_relax: `bool`, optional.
            Allow non-standard WCS keys.
            Default: `True`
        no_fits_standard_units: `bool`, optional
            Skip FITS units standard for units. If this options is choose,
            the units will be printed in header as `~astropy.units.Unit`
            compatible string.
            Default: `False`

        Returns
        -------
        `~astropy.fits.HDUList` :
            HDU storing all FrameData informations.
        """
        return _to_hdu(self, **kwargs)

    def __del__(self):
        """Safe destruction of the container."""
        # ensure all files are removed when exit
        self.disable_memmap()
        # remove tmp folder if empty
        try:
            if len(os.listdir(self.cache_folder)) == 0:
                os.rmdir(self.cache_folder)
        except FileNotFoundError:
            pass

    def to_ccddata(self):
        """Convert actual FrameData to CCDData.

        Returns
        -------
        `~astropy.nddata.CCDData` :
            CCDData instance with actual FrameData informations.
        """
        return _to_ccddata(self)

    def write(self, filename, overwrite=False, **kwargs):
        """Write frame to a fits file."""
        _write_fits(self, filename, overwrite, **kwargs)

    def __copy__(self):
        """Copy the current instance to a new one."""
        return self.copy()

    def median(self, **kwargs):
        """Compute and return the median of the data."""
        med = np.median(self._data, **kwargs)
        if self._unit is None:
            return med
        return med * self.unit

    def mean(self, **kwargs):
        """Compute and return the mean of the data."""
        if self._unit is None:
            return self._data.mean(**kwargs)
        return self._data.mean(**kwargs) * self.unit

    def std(self, **kwargs):
        """Compute and return the std dev of the data."""
        if self._unit is None:
            return self._data.std(**kwargs)
        return self._data.std(**kwargs) * self.unit

    def min(self):
        """Compute minimum value of the data."""
        if self._unit is None:
            return self._data.min()
        return self._data.min() * self.unit

    def max(self):
        """Compute minimum value of the data."""
        if self._unit is None:
            return self._data.max()
        return self._data.max() * self.unit

    def statistics(self):
        """Compute general statistics ofthe image."""
        return {'min': self.min(),
                'max': self.max(),
                'mean': self.mean(),
                'median': self.median(),
                'std': self.std()}



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np
from ..framedata import MemMapArray


def _sep_fix_byte_order(data):
    """Fix the byte order of the data for SEP."""
    if isinstance(data, MemMapArray):
        data = np.array(data)

    if not data.flags['C_CONTIGUOUS']:
        data = np.ascontiguousarray(data)
    if not data.dtype.isnative:
        data = data.byteswap().newbyteorder()
    if data.dtype.type in [np.uint, np.uintc, np.uint8, np.uint16, np.uint32,
                           np.int_, np.intc, np.int8, np.int16, np.int32]:
        data = data.astype(np.float_)
    return data



import sys, os, re
import sphinx_rtd_theme

if not os.path.exists('api'):
    os.mkdir('api')

# Minimum version, enforced by sphinx
needs_sphinx = '2.2.0'

extensions = [
    # 'sphinx.ext.autodoc',
    'sphinxcontrib.napoleon',  # use napoleon to compatibility
    'sphinx.ext.intersphinx',
    'sphinx.ext.coverage',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.graphviz',
    'sphinx.ext.ifconfig',
    'sphinx.ext.todo',
    'matplotlib.sphinxext.plot_directive',
    'IPython.sphinxext.ipython_console_highlighting',
    'IPython.sphinxext.ipython_directive',
    'sphinx.ext.imgmath',
    'nbsphinx',
]

napoleon_use_rtype = False
napoleon_use_ivar = True

nbsphinx_allow_errors = True
nbsphinx_execute = 'never'
nbsphinx_kernel_name = 'python3'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']
# The suffix of source filenames.
source_suffix = '.rst'
# Doc entrypoint
master_doc = 'index'

# Project info
project = 'astropop'
copyright = '2018-2021, Julio Campagnolo and contributors'
import astropop
version = astropop.__version__.split('-', 1)[0]
# The full version, including alpha/beta/rc tags.
release = astropop.__version__
today_fmt = '%B %d, %Y'

# General theme infos
exclude_patterns = ['_templates', '_build', 'Thumbs.db', '.DS_Store']
pygments_style = 'sphinx'
html_theme = "sphinx_rtd_theme"
# html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
html_static_path = ['_static']
imgmath_image_format = 'svg'
htmlhelp_basename = 'astropop'

autosummary_generate = True

default_role = 'py:obj'
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "numpy": ("https://docs.scipy.org/doc/numpy/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/reference", None),
    "matplotlib": ("https://matplotlib.org/", None),
    "astropy": ('http://docs.astropy.org/en/latest/', None)
}



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Small module for simple matrix works. Possible deprecated in future.
"""


import numpy as np


__all__ = ['xy2r', 'iraf_indices', 'trim_array', 'all_equal']


def xy2r(x, y, data, xc, yc):
    """Convert (x, y) values to distance of a (xc, yc) position."""
    r = np.hypot((x-xc), (y-yc))
    return np.ravel(r), np.ravel(data)


def trim_array(data, box_size, position, indices=None, origin=0):
    """Trim a numpy array around a position."""
    x, y = position
    # Correct for 1-based indexing
    x += origin
    y += origin

    dx = dy = float(box_size)/2

    x_min = max(int(x-dx), 0)
    x_max = min(int(x+dx)+1, data.shape[1])
    y_min = max(int(y-dy), 0)
    y_max = min(int(y+dy)+1, data.shape[0])

    d = data[y_min:y_max, x_min:x_max]

    if indices is None:
        return d, x-x_min, y-y_min

    xi = indices[1][y_min:y_max, x_min:x_max]
    yi = indices[0][y_min:y_max, x_min:x_max]
    return d, xi, yi


def all_equal(data):
    """Check if all elements of an array are equal."""
    return np.all(data == data.ravel()[0])


def iraf_indices(data):
    """Create (x, y) index arrays from a data matrix using IRAF convention.

    Iraf convention means (0, 0) on the center of bottom-left pixel.
    """
    y, x = np.indices(data.shape)
    # Matches (0, 0) to the center of the pixel
    # FIXME: Check carefully if this is needed
    # x = x - 0.5
    # y = y - 0.5
    return x, y



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os
import numpy as np
import astroscrappy

from ..logger import logger
from .imarith import imarith
from ..math.physical import convert_to_qfloat
from ..framedata import check_framedata


__all__ = ['cosmics_lacosmic', 'gain_correct', 'subtract_bias',
           'subtract_dark', 'flat_correct', 'process_image',
           'trim_image']


# TODO: replace ccdproc functions by built-in, skiping units
# block_reduce = ccdproc.block_reduce
# block_replicate = ccdproc.block_replicate


def cosmics_lacosmic(frame, inplace=False, **lacosmic_kwargs):
    """
    Remove cosmic rays with LAcosmic. From astroscrappy package.

    Notes
    -----
    * Detailed information about the Laplacian Cosmic Ray Identification method
      can be found on Dokkum,P.G. (2001) - PASP 113, 1420 (2001)
      https://arxiv.org/pdf/astro-ph/0108003.pdf.

    * If `frame` is not a `~astropop.framedata.FrameData` instance,
      inplies in `inplace=False`, and a new `~astropop.framedata.FrameData`
      instance will be created.

    Parameters
    ----------
    frame: `~astropop.framedata.FrameData` compatible
        Values to perform the operation. `~astropy.units.Quantity`, numerical
        values and `~astropy.nddata.CCDData` are also suported.
    inplace: bool, optional
        If True, the operations will be performed inplace in the `frame`.
    logger: `~logging.Logger`
        Python logger to log the actions.

    Returns
    -------
    `~astropop.framedata.FrameData`
        New cosmic-rays corrected `FrameData` instance if not `inplace`,
        else the `image` `~astropop.framedata.FrameData` instance.
    """
    # As lacosmic removes and replace the cosmics pixels, no need to
    # update the mask
    _, dat = astroscrappy.detect_cosmics(frame.data, **lacosmic_kwargs)

    if inplace:
        ccd = frame
    else:
        ccd = frame.copy()

    ccd.data = dat
    # Do not update mask, astroscrappy replace the pixels
    # ccd.mask &= mask
    ccd.header['HIERARCH astropop lacosmic'] = True
    return ccd


def gain_correct(image, gain, inplace=False):
    """
    Process the gain correction of an image.

    Notes
    -----
    * The gain is implemented as a multiplier of the original image.

    * If ``image`` is not a `~astropop.framedata.FrameData` instance,
      inplies in ``inplace=False``, and a new `~astropop.framedata.FrameData`
      instance will be created.

    Parameters
    ----------
    image : `~astropop.framedata.FrameData` compatible
        Values to perform the operation. `~astropy.units.Quantity`, numerical
        values and `~astropy.nddata.CCDData` are also suported.
    gain : float, `~astropy.units.Quantity` or `~astropop.math.QFloat`
        Gain to be applied on the image numerical values.
    inplace : bool, optional
        If True, the operations will be performed inplace in the `image`.

    Returns
    -------
    `~astropop.framedata.FrameData`:
        New gain corrected `FrameData` instance if not ``inplace``, else the
        ``image`` `~astropop.framedata.FrameData` instance.
    """
    gain = convert_to_qfloat(gain)
    nim = imarith(image, gain, '*', inplace=inplace)
    nim.header['HIERARCH astropop gain_corrected'] = True
    nim.header['HIERARCH astropop gain_corrected_value'] = gain.nominal
    nim.header['HIERARCH astropop gain_corrected_unit'] = str(gain.unit)

    return nim


def subtract_bias(image, master_bias, inplace=False):
    """
    Subtract a master_bias frame from a FrameData.

    Notes
    -----
    * This function will just subtract a master bias image from the original
      image. The master bias calculation, by handling several bias images,
      must be done previously.

    * If ``image`` is not a `~astropop.framedata.FrameData` instance,
      inplies in ``inplace=False``, and a new `~astropop.framedata.FrameData`
      instance will be created.

    Parameters
    ----------
    image: `~astropop.framedata.FrameData` compatible
        Image to perform the bias correction. `~astropy.units.Quantity`,
        numerical values and `~astropy.nddata.CCDData` are also suported.
    master_bias: `~astropop.framedata.FrameData` compatible
        Master bias image to be subtracted from the ``image``.
    inplace: bool, optional
        If True, the operations will be performed inplace in the ``image``.

    Returns
    -------
    `~astropop.framedata.FrameData`:
        New bias corrrected `FrameData` instance if ``inplace``, else the
        ``image`` `~astropop.framedata.FrameData` instance.
    """
    master_bias = check_framedata(master_bias)
    nim = imarith(image, master_bias, '-', inplace=inplace)

    nim.header['HIERARCH astropop bias_corrected'] = True
    name = master_bias.origin_filename
    if name is not None:
        nim.header['HIERARCH astropop bias_corrected_file'] = name

    return nim


def subtract_dark(image, master_dark, dark_exposure, image_exposure,
                  inplace=False):
    """
    Subtract master_dark frame from a FrameData.

    Notes
    -----
    * This function will just subtract a master dark frame from the original
      image. The master dark calculation, by handling several dark images,
      must be done previously.

    * Different exposure times among dark image and `FrameData` are handled by
      a multiplying the Master dark image by ``image_exposure/dark_exposure``

    * If ``image`` is not a `~astropop.framedata.FrameData` instance,
      inplies in ``inplace=False``, and a new `~astropop.framedata.FrameData`
      instance will be created.

    Parameters
    ----------
    image : `~astropop.framedata.FrameData` compatible
        Image to perform the dark correction. `~astropy.units.Quantity`,
        numerical values and `~astropy.nddata.CCDData` are also suported.
    master_dark : `~astropop.framedata.FrameData` compatible
        Master dark image to be subtracted from the ``image``.
    dark_exposure : float
        Exposure time of the Master dark.
    image_exposure : float
        Exposure time of the `image`.
    inplace : bool, optional
        If True, the operations will be performed inplace in the `image`.

    Returns
    -------
    `~astropop.framedata.FrameData`:
        New dark corrrected `FrameData` instance if ``inplace``, else the
        ``image`` `~astropop.framedata.FrameData` instance.
    """
    image = check_framedata(image)
    master_dark = check_framedata(master_dark)
    scale = image_exposure/dark_exposure
    if scale != 1:
        logger.debug('Scaling dark by %s factor to match image'
                     ' exposure.', scale)
        master_dark = imarith(master_dark, scale, "*", inplace=False)

    nim = imarith(image, master_dark, '-', inplace=inplace)

    nim.header['HIERARCH astropop dark_corrected'] = True
    nim.header['HIERARCH astropop dark_corrected_scale'] = scale
    name = master_dark.origin_filename
    if name is not None:
        name = os.path.basename(name)
        nim.header['HIERARCH astropop dark_corrected_file'] = name

    return nim


def flat_correct(image, master_flat, min_value=None, norm_value=None,
                 inplace=False):
    """
    Divide the image by a flat field frame.

    Parameters
    ----------
    image : `~astropop.framedata.FrameData` compatible
        Image to perform the flat field correction. `~astropy.units.Quantity`,
        numerical values and `~astropy.nddata.CCDData` are also suported.
    master_flat : `~astropop.framedata.FrameData` compatible
        Master flat field image to be subtracted from the ``image``.
    min_value : float, optional
    norm_value : float, optional
    inplace : bool, optional
        If True, the operations will be performed inplace in the `image`.

    Returns
    -------
    `~astropop.framedata.FrameData`:
        New flat field corrected `FrameData` instance if ``inplace``, else the
        ``image`` `~astropop.framedata.FrameData` instance.
    """
    master_flat = check_framedata(master_flat, copy=True)
    image = check_framedata(image)

    if min_value is not None:
        logger.debug('Set lower flat value to %s', min_value)
        mask = master_flat.data < min_value
        master_flat.data[np.where(mask)] = min_value

    if norm_value is not None:
        logger.debug('Normalizing flat with %s value.', norm_value)
        master_flat = imarith(master_flat, norm_value, '/', inplace=False)

    nim = imarith(image, master_flat, '/', inplace=inplace)

    nim.header['HIERARCH astropop flat_corrected'] = True

    name = master_flat.origin_filename
    if name is not None:
        name = os.path.basename(name)
        nim.header['HIERARCH astropop flat_corrected_file'] = name

    return nim


def trim_image(image, x_slice=None, y_slice=None, inplace=False):
    """Trim an image to a given section. Uses python slice standard.

    Parameters
    ----------
    image : `~astropop.framedata.FrameData` compatible
        Image to be trimmed. `~astropy.units.Quantity`, numerical values and
        `~astropy.nddata.CCDData` are also suported.
    x_slice, y_slice : `slice`
        Section to be trimmed in x and y axes.
    inplace : bool, optional
        If True, the operations will be performed inplace in the `image`.

    Returns
    -------
    `~astropop.framedata.FrameData`:
        Trimmed `FrameData` instance.
    """
    image = check_framedata(image, copy=False)
    if not inplace:
        image = image.copy()

    # shape is always (y, x)
    x_slice = slice(0, image.shape[1]) if x_slice is None else x_slice
    y_slice = slice(0, image.shape[0]) if y_slice is None else y_slice

    section = (y_slice, x_slice)

    # trim the arrays
    data, uncertainty, mask = (image.data, image.get_uncertainty(False),
                               image.mask)
    image.data = data[section]
    image.uncertainty = uncertainty[section]
    if not mask.empty:
        image.mask = mask[section]

    # fix WCS if existing
    if image.wcs is not None:
        wcs = image.wcs.copy()
        wcs.wcs.crpix[0] -= x_slice.start
        wcs.wcs.crpix[1] -= y_slice.start
        # FIXME: this should work but is getting wrong results
        # image.wcs = wcs.slice(section[::-1])
        image.wcs = wcs

    str_slice = ''
    for s in section[::-1]:
        str_slice += ':'.join(str(i) if i is not None else ''
                              for i in [s.start, s.stop])
        str_slice += ','
    str_slice = str_slice[:-1]
    image.meta['HIERARCH astropop trimmed_section'] = str_slice

    return image


def process_image(framedata, master_bias=None, master_dark=None,
                  master_flat=None, gain=None, image_exposure=None,
                  dark_exposure=None, trim=None,
                  lacosmic=False, rebin_func=np.sum,
                  rebin_size=None, readnoise=None, badpixmask=None,
                  overscan=None):
    """Process all the default steps of CCD calibration."""
    raise NotImplementedError



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Reduce polarimetry data."""

from .dualbeam import match_pairs, estimate_dxdy, quarterwave_model, \
                      halfwave_model

__all__ = ['match_pairs', 'estimate_dxdy',
           'quarterwave_model', 'halfwave_model']
